from skystar.core import Variable
from skystar.core import Function
from skystar.core import as_scalar
from skystar.core import as_variable
from skystar.core import config_using
from skystar.core import setup_variable
from skystar.core import no_grad
from skystar.core import TrainingMode,Set_TrainingMode,Get_TrainingMode
from skystar.dataloader import Dataloader
from skystar import utils
from skystar import cuda
from skystar import model
setup_variable()


from PIL import Image
import numpy as np
import os

def init_dir():#初始化一个储存数据的文件夹
    dir_path=os.getcwd()
    dir_path=os.path.join(dir_path,'skystar_dataset')
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)
        os.mkdir(os.path.join(dir_path,'data_txt'))
        print('Dir skystar_dataset created successfully! Please place the .txt file in the data_txt directory')
    return dir_path

def get_names(path):
    '''从txt文本中获取图像的路径'''
    dir_path=init_dir()
    if not os.path.exists(path):
        path=os.path.join(dir_path,'data_txt',path)
        if not os.path.exists(path):
            print('The path {} does not exist!Please place the .txt file in the data_txt directory'.format(path))
            raise FileNotFoundError

    with open(path, 'r', encoding='utf-8') as f:
        dataset = f.readlines()
    train_names = []
    label_names = []
    for name in dataset:
        name = name.split()
        mid = len(name) // 2
        name1, name2 = '', ''
        for i in range(mid):
            if i != mid - 1:
                name1 += (name[i] + ' ')
                name2 += (name[i + mid] + ' ')
            else:
                name1 += name[i]
                name2 += name[i + mid]
        train_names.append(name1)
        label_names.append(name2)
    return train_names, label_names
def image_generator(path):
    '''生成器，按需从路径中加载图片并生成图像数组和标签'''
    names1,names2=get_names(path)
    for i in range(len(names1)):
        name1 = names1[i]
        name2 = names2[i]
        img1 = Image.open(name1)
        img_array1 = np.array(img1)
        img_array1 = img_array1.transpose(2, 0, 1)# 将通道维度放在最前面

        img2 = Image.open(name2)
        img_array2 = np.array(img2)
        yield img_array1, img_array2

def create_dataset(path,shuffle=True):
    '''该函数返回训练数据集与标签数据集，shuffle指定是否打乱数据集的排序，默认打乱'''
    x_dataset,t_dataset=[],[]
    for data1,data2 in image_generator(path):
        x_dataset.append(data1)
        t_dataset.append(data2)
    if shuffle:
        index = np.random.permutation(len(x_dataset))
        x_dataset =np.array([x_dataset[i] for i in index])
        t_dataset=np.array([t_dataset[i] for i in index])
    batch,H,W=t_dataset.shape
    t_dataset=t_dataset.reshape(batch,1,H,W)
    return x_dataset,t_dataset

def data_to_npz(data_txt_path,outfile_name):
    '''使用该函数直接将数据集保存为.npz格式'''
    print('Data Craating......')
    x_dataset,t_dataset=create_dataset(data_txt_path)
    if not os.path.isdir(outfile_name):
        dir_path=init_dir()
        outfile_name=os.path.join(dir_path,outfile_name)
    np.savez_compressed(outfile_name,x_dataset=x_dataset,t_dataset=t_dataset)
    print('Data Craating......Done!')
    return

if __name__ == '__main__':
    data_to_npz('dataset.txt','dataset.npz')


import skystar.utils
import numpy as np
from skystar import no_grad, Get_TrainingMode, Set_TrainingMode
from skystar.core import sigmoid, dropout, ReLU
from skystar.layer import Layer, Gemm, Affine, Convolution, Pooling, BatchNorm, RNN, ResidualBlock, \
    Transpose_Convolution, LSTM, TransitionBlock, DenseBlock, CopyAndCrop, Encoder, Decoder
from skystar.utils import plot_dot_graph
from skystar.optimizer import Adam,SGD,MomentumSGD,AdaGrad
from tqdm import tqdm  # 添加进度条
import time

# =============================================================================
'''model类（继承自Layer）'''
# =============================================================================
class Model(Layer):
    def plot(self, *inputs):
        y = self.forward(*inputs)
        plot_dot_graph(y)

    def predict(self, *inputs):
        y = self.forward(*inputs)
        return y

    def check(self):
        for layer in self._layers:
            print('LayerAttribute: '+layer)
    def __repr__(self):
        info='----------'+self.__class__.__name__+'----------\n'
        for layername in self._layers:
            layer=self.__dict__[layername]
            info+=layer.__repr__()+'\n'
        return info

    def accuracy(self, x_test, t):
        with no_grad():
            y = self.predict(x_test)
            if t.ndim == 2:
                t = np.argmax(t, axis=1)
            y.data = np.argmax(y.data, axis=1)
            sum = t.size
            return np.sum(y.data == t) / sum

    def train(self, train, test=None, mode='Segmentation', lr=0.001, epoch=300, use_gpu=True, interval_epoch=25, plot=True):
        if mode == 'Sort':
            self._train(train=train, test=test, lr=lr, epoch=epoch, plot=False, interval_epoch=interval_epoch,
                        loss_func=skystar.core.softmaxwithloss
                        , accuracy=skystar.utils.accuracy, optimizer=Adam, use_gpu=use_gpu, save_model=True)
        elif mode == 'Segmentation':
            self._train(train=train, test=test, lr=lr, epoch=epoch, plot=plot, interval_epoch=interval_epoch,
                        loss_func=skystar.core.softmaxwithloss
                        , accuracy=skystar.utils.mean_IoU, optimizer=Adam, use_gpu=use_gpu, save_model=True)
        elif mode == 'Sequence':
            self.TrainForSeq(train=train, test=None, lr=lr, epoch=epoch, BPTT=30,
                             loss_func=skystar.core.mean_squared_error,
                             optimizer=Adam, use_gpu=use_gpu, save_model=True)
        else:
            raise ValueError('mode must be "Sort" or "Segmentation" or "Sequence"')
    def _train(self, train, test=None, lr=0.01, epoch=300, plot=False, interval_epoch=30,
               loss_func=skystar.core.softmaxwithloss,
               accuracy=skystar.utils.accuracy, optimizer=Adam, use_gpu=True, save_model=True):
        '''
        :param train: 训练数据迭代器，需要dataloader
        :param interval_epoch: epoch间隔，每隔该轮次保存一次模型以及生成图像
        :param lr: 学习率，默认为0.001
        :param epoch: 学习轮次
        :param test: 测试数据迭代器，默认为None
        :param loss_func: 指定损失函数，默认为softmaxwithloss
        :param accuracy: 指定计算准确率的函数，默认为工具中的accuracy
        :param plot: 是否在过程中画图，画图功能只有在test中才可用
        :param optimizer: 优化器，默认为Adam
        :param use_gpu: 是否使用gpu，默认为Ture
        :return:
        '''
        print('===============Training Begin===============')
        if use_gpu:
            if skystar.cuda.gpu_enable:
                self.to_gpu()
                train.to_gpu()
                print('GPU is available, and all parameters are converted to cp.ndarray')
                if test is not None:
                    test.to_gpu()
            else:
                use_gpu = False
        self.results = [['train_loss'], ['train_acc'], ['test_acc']]
        optimizer = optimizer(lr).setup(self)
        for i in range(epoch):
            if not Get_TrainingMode():
                Set_TrainingMode(True)
            print(f'Epoch {i + 1}:')
            sum_acc,sum_loss = 0.0,0.0
            time.sleep(0.01)  # 停顿0.1秒，避免输出条与输出字符串同步
            # 使用 tqdm 包裹训练数据集
            for x, t in tqdm(train, desc='Training', total=train.max_iter):
                y = self(x)
                loss = loss_func(y, t)
                self.cleangrads()
                loss.backward()
                optimizer.update()
                if accuracy is not None:
                    sum_acc += accuracy(y, t) * len(t)
                sum_loss += loss.data

            print(f'Train_Loss {sum_loss / train.data_size}')
            self.results[0].append(sum_loss / train.data_size)
            if accuracy is not None:
                print(f'Train_Acc {sum_acc / train.data_size}')
                self.results[1].append(sum_acc / train.data_size)

            if test is not None:
                '''这里已经将数据加载到gpu上，因此默认use_gpu为False'''
                if ((i+1) % interval_epoch == 0 or i==epoch-1) and plot:
                    self.results[2].append(self.eval(test, use_gpu=False, accuracy=accuracy, plot=True, plot_num=2, label=f'Epoch{i + 1}_'))
                else:
                    self.results[2].append(self.eval(test, use_gpu=False, accuracy=accuracy, plot=False, plot_num=2))

            '''自动保存模型'''
            if save_model and ((i+1)%interval_epoch==0 or i==epoch-1):
                name = self.__class__.__name__
                print('Saving the model params......')
                self.save_weights(filename=name)
                print('Saving the results......')
                for result in self.results:
                    if result:
                        name=result[0]+'.txt'
                        skystar.utils.write_text(result[1:],name)
                if use_gpu:
                    self.to_gpu()

        if save_model:
            print('Saving the model onnx......')
            self.save_to_onnx(x)
            if use_gpu:
                self.to_gpu()


    def eval(self, test, use_gpu=True, accuracy=skystar.utils.accuracy,plot=False,plot_num=2,label=''):
        '''评估模式下，如果使用plot=ture，'''
        Set_TrainingMode(False)
        print('===============Test Begin===============')
        if use_gpu:
            if skystar.cuda.gpu_enable:
                self.to_gpu()
                test.to_gpu()
                print('GPU is available, and all parameters are converted to cp.ndarray')
        plot_interval_num=test.max_iter//plot_num
        iter_num=0
        sum_acc = 0.0
        with skystar.core.no_grad():
            for x, t in tqdm(test, desc='Testing', total=test.max_iter):
                iter_num+=1
                y = self(x)
                if accuracy is not None:
                    sum_acc += accuracy(y, t) * len(t)
                if iter_num%plot_interval_num==0 and plot:
                    predict=y.data.argmax(axis=1)
                    predict=predict.reshape(predict.shape[0],1,predict.shape[1],predict.shape[2])
                    img=skystar.utils.Create_ImgForSeg((x,t,predict))
                    skystar.utils.save_img(img,name=label+f'num{iter_num}')

        print(f'Test_Acc: {sum_acc / len(test)}')
        return sum_acc / len(test)

    def TrainForSeq(self, train, test=None, lr=0.001, epoch=300, BPTT=30,loss_func=skystar.core.mean_squared_error,
                    optimizer=Adam, use_gpu=True, save_model=True):
        '''专门用于序列模型的训练函数'''
        Set_TrainingMode(True)
        print('===============Training Begin===============')
        if use_gpu:
            print('GPU is available, and all parameters are converted to cp.ndarray')
            if skystar.cuda.gpu_enable:
                self.to_gpu()
                train.to_gpu()
                if test is not None:
                    test.to_gpu()
        self.list = [[], [], []]
        optimizer = optimizer(lr).setup(self)
        for i in range(epoch):#训练开始
            self.reset_state()
            print(f'Epoch {i + 1}:')
            loss,count = 0.0, 0
            for x, t in tqdm(train, desc='Training', total=train.max_iter):
                y=self(x)
                loss+=loss_func(y, t)
                count+=1
                if count%BPTT == 0 or count == train.data_size:
                    self.cleangrads()
                    loss.backward()
                    loss.unchain_backward()
                    optimizer.update()
            avg_loss = loss / count
            self.list[0].append(avg_loss)
            print(f'Train_Loss {avg_loss}')
        if save_model:
            name = self.__class__.__name__
            print('Saving the model params......')
            self.save_weights(filename=name)

# =============================================================================
'''一些经典的model'''
# =============================================================================
class MLP(Model):
    '''fc_output_sizes：tuple，输入层的神经元个数，全连接层神经网络'''

    def __init__(self, fc_output_sizes, activation=sigmoid):
        super().__init__()
        self.activation = activation
        self.layers = []

        for i, out_size in enumerate(fc_output_sizes):
            layer = Gemm(out_size)
            setattr(self, 'l' + str(i), layer)
            self.layers.append(layer)

    def forward(self, x):
        for l in self.layers[:-1]:
            x = self.activation(l(x))
        return self.layers[-1](x)


class Batch_norm_MLP(Model):
    '''添加了batchnorm的全连接层，分测试和训练阶段'''

    def __init__(self, fc_output_sizes, activation=sigmoid, gamma=1.0, beta=0, momentum=0.9):
        super().__init__()
        self.activation = activation
        self.layers = []
        self.gamma = gamma
        self.beta = beta
        self.momentum = momentum

        for i, outsize in enumerate(fc_output_sizes):
            layer1 = Affine(outsize)
            setattr(self, 'Affine' + str(i), layer1)
            self.layers.append(layer1)
            layer2 = BatchNorm(self.gamma, self.beta, self.momentum)
            setattr(self, 'Batch_norm' + str(i), layer2)
            self.layers.append(layer2)

    def forward(self, x):
        for i in range(0, len(self.layers) - 1):
            layer = self.layers[i]
            x = layer(x)
            if i % 2 != 0:
                x = self.activation(x)

        layer = self.layers[-1]
        x = layer(x)
        return x


class Batchnorm_dropout_MLP(Model):
    '''添加了batchnorm的dropout的全连接层，分测试和训练阶段,dropout层在sigmoid后面'''

    def __init__(self, fc_output_sizes, activation=sigmoid, use_dropout=False, dropout_ratio=0.5, gamma=1.0, beta=0,
                 momentum=0.9):
        super().__init__()
        self.activation = activation
        self.layers = []
        self.use_dropout = use_dropout
        self.dropout_ratio = dropout_ratio
        self.gamma = gamma
        self.beta = beta
        self.momentum = momentum

        for i, outsize in enumerate(fc_output_sizes):
            layer1 = Affine(outsize)
            setattr(self, 'Affine' + str(i), layer1)
            self.layers.append(layer1)
            layer2 = BatchNorm(self.gamma, self.beta, self.momentum)
            setattr(self, 'Batch_norm' + str(i), layer2)
            self.layers.append(layer2)

    def forward(self, x):
        for i in range(0, len(self.layers) - 1):
            layer = self.layers[i]
            x = layer(x)
            if i % 2 != 0:
                x = self.activation(x)
                if self.use_dropout:
                    x = dropout(x)

        layer = self.layers[-1]
        x = layer(x)
        return x


class Simple_CNN(Model):
    def __init__(self, activation=ReLU):
        super().__init__()
        self.activation = activation
        self.convolution1 = Convolution(16, 3, 3)
        self.pooling1 = Pooling(2, stride=2)
        self.convolution2 = Convolution(32, 3, 3)
        self.pooling2 = Pooling(2, stride=2)
        self.affine = Gemm(10)

    def forward(self, x):
        y1 = self.convolution1(x)
        y1 = self.activation(y1)
        y1 = self.pooling1(y1)

        y2 = self.convolution2(y1)
        y2 = self.activation(y2)
        y2 = self.pooling2(y2)

        out = self.affine(y2)
        return out


class Simple_RNN(Model):
    def __init__(self, hidden_size, out_size):
        super().__init__()
        self.RNN = RNN(hidden_size=hidden_size)  # 输出状态
        self.affine = Affine(out_size)  # 使用affine层输出状态

    def reset_state(self):
        self.RNN.reset_state()

    def forward(self, x):
        x = self.RNN(x)
        x = self.affine(x)

        return x


class Better_RNN(Model):
    def __init__(self, hidden_size, out_size):
        super().__init__()
        self.RNN = LSTM(hidden_size=hidden_size)
        self.affine = Affine(out_size)

    def reset_state(self):
        self.RNN.reset_state()

    def forward(self, x):
        x = self.RNN(x)
        x = self.affine(x)

        return x


class VGG(Model):
    def __init__(self, output=1000, ratio=0.5):
        super().__init__()
        self.ratio = ratio
        self.conv1_1 = Convolution(16, 3, 3, pad=1)
        self.conv1_2 = Convolution(16, 3, 3, pad=1)
        self.pool1 = Pooling(2, stride=2)
        self.conv2_1 = Convolution(128, 3, 3, pad=1)
        self.conv2_2 = Convolution(128, 3, 3, pad=1)
        self.pool2 = Pooling(2, stride=2)
        self.conv3_1 = Convolution(256, 3, 3, pad=1)
        self.conv3_2 = Convolution(256, 3, 3, pad=1)
        self.conv3_3 = Convolution(256, 3, 3, pad=1)
        self.pool3 = Pooling(2, stride=2)
        self.conv4_1 = Convolution(512, 3, 3, pad=1)
        self.conv4_2 = Convolution(512, 3, 3, pad=1)
        self.conv4_3 = Convolution(512, 3, 3, pad=1)
        self.pool4 = Pooling(2, stride=2)
        self.conv5_1 = Convolution(512, 3, 3, pad=1)
        self.conv5_2 = Convolution(512, 3, 3, pad=1)
        self.conv5_3 = Convolution(512, 3, 3, pad=1)
        self.pool5 = Pooling(2, stride=2)
        self.affine1 = Affine(4096)
        self.affine2 = Affine(4096)
        self.affine3 = Affine(output)

    def forward(self, x):
        '''激活函数用ReLU'''
        x = ReLU(self.conv1_1(x))
        x = ReLU(self.conv1_2(x))
        x = self.pool1(x)
        x = ReLU(self.conv2_1(x))
        x = ReLU(self.conv2_2(x))
        x = self.pool2(x)
        x = ReLU(self.conv3_1(x))
        x = ReLU(self.conv3_2(x))
        x = ReLU(self.conv3_3(x))
        x = self.pool3(x)
        x = ReLU(self.conv4_1(x))
        x = ReLU(self.conv4_2(x))
        x = ReLU(self.conv4_3(x))
        x = self.pool4(x)
        x = ReLU(self.conv5_1(x))
        x = ReLU(self.conv5_2(x))
        x = ReLU(self.conv5_3(x))
        x = self.pool5(x)
        x = ReLU(self.affine1(x))
        x = dropout(x, ratio=self.ratio)
        x = ReLU(self.affine2(x))
        x = dropout(x, ratio=self.ratio)
        x = self.affine3(x)

        return x


class STL_10_CNN(Model):
    def __init__(self, ratio=0.5, output_size=10):
        super().__init__()
        self.ratio = ratio
        self.conv1_1 = Convolution(16, 3, 3, pad=1)
        self.conv1_2 = Convolution(16, 3, 3, pad=1)
        self.pool1 = Pooling(2, stride=2)
        self.conv2_1 = Convolution(128, 3, 3, pad=1)
        self.conv2_2 = Convolution(128, 3, 3, pad=1)
        self.pool2 = Pooling(2, stride=2)
        self.conv3_1 = Convolution(256, 3, 3, pad=1)
        self.conv3_2 = Convolution(256, 3, 3, pad=1)
        self.pool3 = Pooling(2, stride=2)
        # self.conv4_1=Convolution(512,3,3,pad=1)
        # self.conv4_2=Convolution(512,3,3,pad=1)
        # self.pool4=Pooling(2,stride=2)
        # self.conv5_1=Convolution(512,3,3,pad=1)
        # self.conv5_2=Convolution(512,3,3,pad=1)
        # self.pool5=Pooling(2,stride=2)
        self.affine1 = Affine(100)
        self.affine2 = Affine(100)
        self.affine3 = Affine(output_size)

    def forward(self, x):
        '''激活函数用ReLU'''
        x = ReLU(self.conv1_1(x))
        x = ReLU(self.conv1_2(x))
        x = self.pool1(x)
        x = ReLU(self.conv2_1(x))
        x = ReLU(self.conv2_2(x))
        x = self.pool2(x)
        x = ReLU(self.conv3_1(x))
        x = ReLU(self.conv3_2(x))
        x = self.pool3(x)
        # x=ReLU(self.conv4_1(x))
        # x=ReLU(self.conv4_2(x))
        # x=self.pool4(x)
        # x=ReLU(self.conv5_1(x))
        # x=ReLU(self.conv5_2(x))
        # x=self.pool5(x)
        x = ReLU(self.affine1(x))
        x = dropout(x, ratio=self.ratio)
        x = ReLU(self.affine2(x))
        x = dropout(x, ratio=self.ratio)
        x = self.affine3(x)

        return x


class Simple_FCN(Model):
    def __init__(self, activation=ReLU):
        super().__init__()
        self.activation = activation
        self.conv1_1 = Convolution(16, 3, 3, pad=1)
        self.conv1_2 = Convolution(16, 3, 3, pad=1)
        self.pool1 = Pooling(2, stride=2)
        self.conv2_1 = Convolution(32, 3, 3, pad=1)
        self.conv2_2 = Convolution(32, 3, 3, pad=1)
        self.pool2 = Pooling(2, stride=2)
        self.conv3_1 = Convolution(64, 3, 3, pad=1)
        self.conv3_2 = Convolution(64, 3, 3, pad=1)
        self.pool3 = Pooling(2, stride=2)
        self.conv_31 = Convolution(2, 1, 1)

        self.conv4_1 = Convolution(128, 3, 3, pad=1)
        self.pool4 = Pooling(2, stride=2)
        self.conv_21 = Convolution(2, 1, 1)

        self.conv5_1 = Convolution(256, 3, 3, pad=1)
        self.pool5 = Pooling(2, stride=2)
        self.conv_11 = Convolution(4096, 1, 1)
        self.conv_12 = Convolution(4096, 1, 1)
        self.conv_13 = Convolution(2, 1, 1)
        self.transposed_conv11 = Transpose_Convolution(2, 4, 4, stride=2, pad=1, nobias=True)
        self.transposed_conv12 = Transpose_Convolution(2, 4, 4, stride=2, pad=1, nobias=True)
        self.transposed_conv13 = Transpose_Convolution(2, 32, 24, stride=8, pad=8, nobias=True)

    def forward(self, x):
        y1 = self.conv1_1(x)
        y1 = self.conv1_2(y1)
        y1 = self.activation(y1)
        y1 = self.pool1(y1)

        y1 = self.conv2_1(y1)
        y1 = self.conv2_2(y1)
        y1 = self.activation(y1)
        y1 = self.pool2(y1)

        y1 = self.conv3_1(y1)
        y1 = self.conv3_2(y1)
        y1 = self.activation(y1)
        y1 = self.pool3(y1)
        y3 = self.conv_31(y1)

        y1 = self.conv4_1(y1)
        y1 = self.activation(y1)
        y1 = self.pool4(y1)
        y2 = self.conv_21(y1)

        y1 = self.conv5_1(y1)
        y1 = self.activation(y1)
        y1 = self.pool5(y1)

        y1 = self.conv_11(y1)
        y1 = self.conv_12(y1)
        y1 = self.conv_13(y1)

        y1 = self.transposed_conv11(y1)
        y1 = y1 + y2
        y1 = self.transposed_conv12(y1)
        y1 = y1 + y3
        y1 = self.transposed_conv13(y1)
        return y1


class Simple_ResNet(Model):
    def __init__(self, activation=ReLU):
        super().__init__()
        self.activation = activation
        self.conv1 = Convolution(16, 3, 3, pad=1)
        self.pool1 = Pooling(2, stride=2)
        self.residual1_1 = ResidualBlock(16, stride=1, use_conv1x1=False)
        self.residual1_2 = ResidualBlock(16, stride=1, use_conv1x1=False)
        self.residual2_1 = ResidualBlock(32, stride=2, use_conv1x1=True)
        self.residual2_2 = ResidualBlock(32, stride=1, use_conv1x1=False)
        self.residual3_1 = ResidualBlock(64, stride=2, use_conv1x1=True)
        self.residual3_2 = ResidualBlock(64, stride=1, use_conv1x1=False)
        self.affine = Affine(10)

    def forward(self, x):
        x = ReLU(self.conv1(x))
        x = self.pool1(x)
        x = self.residual1_1(x)
        x = self.residual1_2(x)
        x = self.residual2_1(x)
        x = self.residual2_2(x)
        x = self.residual3_1(x)
        x = self.residual3_2(x)
        x = self.affine(x)
        return x


class Simple_densenet(Model):
    def __init__(self, activation=ReLU, to_gpu=True):
        super().__init__()
        self.activation = activation
        self.conv1 = Convolution(16, 3, 3, pad=1)
        self.BN1 = BatchNorm()
        self.pool1 = Pooling(2, stride=2)
        self.dense1 = DenseBlock(32, 4)
        self.transition1 = TransitionBlock(32)
        self.dense2 = DenseBlock(64, 4)
        self.transition2 = TransitionBlock(64)
        self.dense3 = DenseBlock(64, 4)
        self.transition3 = TransitionBlock(64)
        self.affine = Affine(10)

    def forward(self, x):
        x = self.conv1(x)
        x = self.BN1(x)
        x = ReLU(x)
        x = self.pool1(x)
        x = self.dense1(x)
        x = self.transition1(x)
        x = self.dense2(x)
        x = self.transition2(x)
        x = self.dense3(x)
        x = self.transition3(x)
        x = self.affine(x)
        return x


class U_net(Model):
    def __init__(self):
        super().__init__()
        self.conv1_1 = Convolution(16, 3, 3)
        self.conv1_2 = Convolution(16, 3, 3)
        self.pool1 = Pooling(2, stride=2)
        self.conv2_1 = Convolution(128, 3, 3)
        self.conv2_2 = Convolution(128, 3, 3)
        self.pool2 = Pooling(2, stride=2)
        self.conv3_1 = Convolution(256, 3, 3)
        self.conv3_2 = Convolution(256, 3, 3)
        self.pool3 = Pooling(2, stride=2)
        self.conv4_1 = Convolution(512, 3, 3)
        self.conv4_2 = Convolution(512, 3, 3)
        self.pool4 = Pooling(2, stride=2)

        self.conv5_1 = Convolution(512, 3, 3)
        self.conv5_2 = Convolution(512, 3, 3)
        self.transconv1 = Transpose_Convolution(512, 2, 2, stride=2, nobias=True)
        self.conv6_1 = Convolution(512, 3, 3)
        self.conv6_2 = Convolution(512, 3, 3)
        self.transconv2 = Transpose_Convolution(256, 2, 2, stride=2, nobias=True)
        self.conv7_1 = Convolution(256, 3, 3)
        self.conv7_2 = Convolution(256, 3, 3)
        self.transconv3 = Transpose_Convolution(128, 2, 2, stride=2, nobias=True)
        self.conv8_1 = Convolution(128, 3, 3)
        self.conv8_2 = Convolution(128, 3, 3)
        self.transconv4 = Transpose_Convolution(16, 2, 2, stride=2, nobias=True)
        self.conv9_1 = Convolution(16, 3, 3)
        self.conv9_2 = Convolution(16, 3, 3)

    def forward(self, x):
        x1 = ReLU(self.conv1_1(x))  # 572
        x1 = ReLU(self.conv1_2(x1))  # 570
        x1_crop = CopyAndCrop((392, 392))(x1)

        x2 = self.pool1(x1)  # 568
        x2 = ReLU(self.conv2_1(x2))  # 284
        x2 = ReLU(self.conv2_2(x2))  # 282
        x2_crop = CopyAndCrop((200, 200))(x2)

        x3 = self.pool2(x2)  # 280
        x3 = ReLU(self.conv3_1(x3))  # 140
        x3 = ReLU(self.conv3_2(x3))  # 138
        x3_crop = CopyAndCrop((104, 104))(x3)

        x4 = self.pool3(x3)  # 136
        x4 = ReLU(self.conv4_1(x4))  # 68
        x4 = ReLU(self.conv4_2(x4))  # 66
        x4_crop = CopyAndCrop((56, 56))(x4)

        x = self.pool4(x4)  # 64
        x = ReLU(self.conv5_1(x))  # 32
        x = ReLU(self.conv5_2(x))  # 30

        x = self.transconv1(x) + x4_crop  # 512 28
        x = ReLU(self.conv6_1(x))  # 512 56
        x = ReLU(self.conv6_2(x))

        x = self.transconv2(x) + x3_crop  # 512 52
        x = ReLU(self.conv7_1(x))  # 256 104
        x = ReLU(self.conv7_2(x))

        x = self.transconv3(x) + x2_crop  # 256 100
        x = ReLU(self.conv8_1(x))  # 128 200
        x = ReLU(self.conv8_2(x))

        x = self.transconv4(x) + x1_crop  # 128 196
        x = ReLU(self.conv9_1(x))  # 16 392
        x = ReLU(self.conv9_2(x))  # 16 390
        return x
class Transformer(Layer):
    def __init__(self,word_num=10,embedding_dim=512,dff=2048, dkv=64, n_heads=8):
        '''
        :param word_num:单词的总数
        :param embedding_dim:词嵌入的维度
        :param dff:前向传播层的隐藏层维度
        :param dkv:自注意力层k，v的维度
        :param n_heads:多头注意力的头数
        '''
        super().__init__()
        self.name='Transformer'
        self.embedding_dim=embedding_dim
        self.Encoder=Encoder(word_num,embedding_dim,dff=dff,dkv=dkv,n_heads=n_heads)
        self.Decoder=Decoder(word_num,embedding_dim,dff=dff,dkv=dkv,n_heads=n_heads)
        self.Gemm=Gemm(word_num)#最后映射输出
    def forward(self,enc_input,dec_input):
        batch,sqrlen=dec_input.shape
        enc_output=self.Encoder(enc_input)
        dec_output=self.Decoder(dec_input,enc_input,enc_output)
        dec_self_attention=self.Decoder.MutiHead.attention
        dec_cross_attention=self.Decoder.CorssMutiHead.attention

        dec_output=dec_output.reshape(-1,self.embedding_dim)
        dec_output=self.Gemm(dec_output)
        dec_output=dec_output.reshape(batch,sqrlen,-1)
        return dec_output,dec_self_attention,dec_cross_attention


import numpy as np
import onnx
try:
    import cupy
    inttype = (cupy.int32, cupy.int64, np.int32, np.int64)
except ImportError:
    inttype = (np.int32, np.int64)
from onnx import TensorProto
from onnx.helper import (make_model, make_node, make_graph,
                         make_tensor, make_tensor_value_info)
from onnx.checker import check_model
from onnxsim import simplify

import skystar


class Node:
    def __init__(self, node, generation):
        self.generation = generation
        self.node = node


class Graph:
    def __init__(self):
        self.nodes = []
        self.inputs = []
        self.outputs = []
        self.initializers = []
        self.inputsinfo=[]
        self.last_func_id = None
        self.inputindex = 1

    def add_node(self, node):
        self.nodes.append(node)

    def set_inputs(self, inputs):
        self.inputs = inputs

    def set_outputs(self, outputs):
        self.outputs = outputs


# ===============================================================
# 将graph保存为onnx
# ===============================================================
def save_graph(graph, model_name, file_name='Example.onnx', ifsimplify=False, version=15):
    _graph = make_graph(
        nodes=graph.nodes,
        name=model_name,
        inputs=graph.inputs,
        outputs=graph.outputs,
        initializer=graph.initializers
    )
    onnx_model = make_model(graph=_graph, opset_imports=[onnx.helper.make_opsetid("", version)])  # 指定版本
    check_model(model=onnx_model)
    if ifsimplify:
        model_simplified, check = simplify(onnx_model)
        assert check, "Simplified ONNX model could not be validated"
        onnx.save(model_simplified, file_name)
        print('model saved and simplified successfully--Path:{}'.format(file_name))
    else:
        onnx.save_model(onnx_model, file_name)
        print('model saved successfully--Path:{}'.format(file_name))


# ===============================================================
# 载入和使用onnx模型,（放弃使用）
# ===============================================================
# def load_model(model_name):
#     session = ort.InferenceSession(model_name)
#     return session
#
#
# def model_predict(model, input):
#     input = input.astype(np.float32)
#     inputs = {"input": input}
#     outputs = model.run('output', inputs)
#     return outputs


# ===============================================================
# 创建graph
# ===============================================================
def create_graph(output):
    graph = Graph()
    if isinstance(output, tuple):
        output = output[0]
    fs = [output.creator]
    graph.last_func_id = id(fs[0])
    graph.outputs.append(make_tensor_value_info('Output', TensorProto.FLOAT, list(output.shape)))
    nodes = []
    while fs:
        f = fs.pop()
        generate_initializers(f, graph)  # 先生成initializers
        graph_node = _graph_node(f, graph)  # 后生成nodes
        if graph_node.node not in nodes:
            graph.nodes.append(graph_node)
            nodes.append(graph_node.node)
            for input in f.inputs:
                if input.creator is not None and input.creator not in fs:
                    fs.append(input.creator)
    graph.nodes = sorted(graph.nodes, key=lambda x: x.generation)
    graph.nodes = [i.node for i in graph.nodes]
    return graph


# ===============================================================
# 根据f的名称调用相应的create函数创建节点
# ===============================================================
def _graph_node(f, graph):
    name = f.__class__.__name__
    if name in function_nodes:
        return function_nodes[name](f, graph)
    else:
        print(f'Warning:No such function node: {name}')


# ===============================================================
# 生成输入输出的名称,输入输出是core的Function函数里前向传播的输入输出，大多数create函数可用
# ===============================================================
def generate_names(f, graph):
    inputs_name = []
    if f.generation != 0:
        for input in f.inputs:
            if input.data is not None:  # 考虑到卷积层的b可能为None
                if input.name is not None:
                    inputs_name.append(input.name + f'_{id(input)}')
                else:
                    inputs_name.append(f'mid_{id(input)}')
    else:
        for input in f.inputs:
            if id(input) in graph.inputsinfo:
                continue
            else:
                if input.name is None:
                    index = len(graph.inputs)+1
                    node = make_tensor_value_info(f'Input{index}', TensorProto.FLOAT, list(input.shape))
                    graph.inputs.append(node)
                    inputs_name.append(f'Input{index}')
                else:
                    inputs_name.append(input.name + f'_{id(input)}')

    outputs_name = [f'mid_{id(f.outputs[0]())}']
    if id(f) == graph.last_func_id:
        outputs_name = ['Output']
    return inputs_name, outputs_name

# ===============================================================
# 根据函数的输入生成initializers，如果函数的输入具有name，则将其纳入initializers
# ===============================================================
def generate_initializers(f, graph):
    for input in f.inputs:
        if input.name is None and input.generation==0 and f.generation != 0:
            input.name = 'mid'#这里是对那些除开输入数据外，首次出现的数据进行命名
        if input.name is not None and input.data is not None:
            if isinstance(input.data, inttype):
                datatype=TensorProto.INT32
            else:
                datatype=TensorProto.FLOAT
            initializer = make_tensor(input.name + f'_{id(input)}', datatype, list(input.shape), input.data)
            if initializer not in graph.initializers:
                graph.initializers.append(initializer)

# ===============================================================
# create函数，创建Function节点
# ===============================================================
def create_node(f, graph, node_type, **kwargs):
    inputs_name, outputs_name = generate_names(f, graph)
    node = make_node(
        node_type,
        inputs_name,
        outputs_name,
        name=f'{node_type}_node_{id(f)}',
        **kwargs
    )
    return Node(node, f.generation)


# ===============================================================
# 创建Function节点
# ===============================================================
def create_add_node(f, graph):
    return create_node(f, graph, 'Add')
def create_matmul_node(f, graph):
    return create_node(f, graph, 'MatMul')
def create_multify_node(f, graph):
    return create_node(f, graph, 'Mul')
def create_relu_node(f, graph):
    return create_node(f, graph, 'Relu')
def create_maxpool_node(f, graph):
    return create_node(f, graph, 'MaxPool', kernel_shape=[f.pool_size, f.pool_size],
                       strides=[f.stride, f.stride], pads=[0, 0, f.pad, f.pad])
def create_AveragePool_node(f, graph):
    return create_node(f, graph, 'AveragePool', kernel_shape=[f.pool_size, f.pool_size],
                       strides=[f.stride, f.stride], pads=[0, 0, f.pad, f.pad])
def create_dropout_node(f, graph):
    return create_node(f, graph, 'Dropout')
def create_conv_node(f, graph):
    '''需要初始化权重'''
    return create_node(f, graph, 'Conv',
                       kernel_shape=[f.inputs[1].shape[2], f.inputs[1].shape[3]],
                       strides=[f.stride, f.stride],
                       pads=[f.pad, f.pad, f.pad, f.pad], )
def create_convtranspose_node(f, graph):
    return create_node(f, graph, 'ConvTranspose', kernel_shape=[f.inputs[1].shape[2], f.inputs[1].shape[3]],
                       strides=[f.stride, f.stride],
                       pads=[f.pad, f.pad, f.pad, f.pad],
                       output_shape=list(f.outputs[0]().shape), )
def create_softmaxcrossentropyloss_node(f, graph):
    '''该节点需要特殊设置'''
    inputs_name = [f'mid_{id(f.inputs[0])}', 'Label']
    graph.inputs.append(make_tensor_value_info('Label', TensorProto.INT64, list(f.inputs[1].shape)))
    f.inputs[1].name = 'Label'
    outputs_name = 'Output'
    node = make_node(
        'SoftmaxCrossEntropyLoss',
        inputs_name,
        outputs_name,
        reduction='mean',
        ignore_index=None,
        name=f'SoftmaxCrossEntropyLoss_node_{id(f)}'
    )
    return Node(node, f.generation)
def create_batchNormalization_node(f, graph):
    return create_node(f, graph, 'BatchNormalization', epsilon=1e-5, momentum=f.momentum, training_mode=0)
def create_sigmoid_node(f, graph):
    return create_node(f, graph, 'Sigmoid')
def create_softmax_node(f, graph):
    return create_node(f, graph, 'Softmax', axis=1)
def create_meansquarederror_node(f, graph):
    pass
def create_gemm_node(f, graph):
    return create_node(f, graph, 'Gemm', alpha=f.alpha, beta=f.beta, transA=f.transA, transB=f.transB)
def create_Concat_node(f, graph):
    return create_node(f, graph, 'Concat', axis=f.axis)
def create_Reshape_node(f, graph):
    return create_node(f,graph,'Reshape')
def create_slice_node(f, graph):
    return create_node(f, graph, 'Slice')
def create_tanh_node(f, graph):
    return create_node(f, graph, 'Tanh')
def create_dic_node(f, graph):
    return create_node(f, graph, 'Div')
def create_transpose_node(f, graph):#矩阵转置,不完善
    return create_node(f, graph, 'Transpose',perm=None)
def create_reducemean_node(f, graph):
    pass
def create_layernorm_node(f,graph):
    return create_node(f, graph, 'LayerNormalization', axis=-1, epsilon=1e-5)
def create_gather_node(f, graph):
    return create_node(f, graph, 'Gather', axis=f.axis)
def create_expand_node(f, graph):
    return create_node(f, graph, 'Expand')
def create_sub_node(f,graph):
    return create_node(f, graph, 'Sub')
# 使用节点的字典
function_nodes = {
    'SoftmaxCrossEntropyLoss': create_softmaxcrossentropyloss_node,
    'Sigmoid': create_sigmoid_node,
    'Relu': create_relu_node,
    'Softmax': create_softmax_node,
    'MeanSquaredError': create_meansquarederror_node,
    'Gemm': create_gemm_node,
    'BatchNormalization': create_batchNormalization_node,
    'Dropout': create_dropout_node,
    'Conv': create_conv_node,
    'ConvTranspose': create_convtranspose_node,
    'MaxPool': create_maxpool_node,
    'Add': create_add_node,
    'Mul': create_multify_node,
    'Dot': create_matmul_node,
    "Concat": create_Concat_node,
    "AveragePool": create_AveragePool_node,
    'Reshape': create_Reshape_node,
    'Slice': create_slice_node,
    'Tanh': create_tanh_node,
    'Div': create_dic_node,
    'Transpose': create_transpose_node,
    'LayerNorm': create_layernorm_node,
    'BroadcastTo':create_expand_node,
    'Gather':create_gather_node,
    'Matmul':create_matmul_node,
    'Sub': create_sub_node,
}



import math
import numpy as np
from skystar import cuda

class Dataloader:
    '''从Dataloader取出的数据维度ndim一致'''
    def __init__(self,dataset,batch_size,shuffle=True,gpu=False,dtype=np.float32):
        self.dataset=dataset
        self.batch_size=batch_size
        self.shuffle=shuffle
        self.data_size=len(dataset)
        self.max_iter=math.ceil(self.data_size/batch_size)#向上取整
        self.gpu=gpu

        self.reset()

    def reset(self):
        self.iteration=0
        if self.shuffle:
            self.index=np.random.permutation(self.data_size)
        else:
            self.index=np.arange(self.data_size)

    def __iter__(self):
        return self

    def __next__(self):
        if self.iteration>=self.max_iter:
            self.reset()
            raise StopIteration

        '''按照batch大小取出数据，如果最后一部分数据少于batch，那么直接取出剩余数据'''
        i,batch_size=self.iteration,self.batch_size
        batch_index=self.index[i*batch_size:(i+1)*batch_size]
        batch=[self.dataset[i] for i in batch_index]
        xp=cuda.cupy if self.gpu else np
        x=xp.array([example[0] for example in batch],dtype=xp.float32)
        t=xp.array([example[1] for example in batch],dtype=xp.int32)

        self.iteration+=1
        return x,t

    def next(self):
        return self.__next__()

    def to_cpu(self):
        self.gpu=False

    def to_gpu(self):
        self.gpu=True

    def __len__(self):
        return self.max_iter



class SeqDataloader(Dataloader):
    '''创建一个用于时间序列模型的数据加载器，如果数据量为1500，batch=3，那么每一小批量数据将分别从[1,501,1001]开始，一直到[500,100,1500]'''
    def __init__(self,dataset,batch_size,gpu=False):
        super().__init__(dataset=dataset,batch_size=batch_size,shuffle=False,gpu=gpu)#时间序列模型训练数据不打乱

    def __next__(self):
        if self.iteration>=self.max_iter:
            self.reset()
            raise StopIteration
        '''偏移：假如一个时序数据（1，2，3，4，5，6），batch为2，那么数据分为两组，取三次数据（1，4）（2，5）（3，6）
        如果数据本身就是一个序列，如（1，4）是两个具有高维的序列数据，偏移与否不影响最后的输出'''
        jump=self.data_size//self.batch_size#获取偏移量,用于时序偏移
        batch_index=[(i*jump+self.iteration)%self.data_size for i in range(self.batch_size)]
        batch=[self.dataset[i] for i in batch_index]
        xp = cuda.cupy if self.gpu else np
        if len(batch[0])==2:
            x=xp.array([example[0] for example in batch],dtype=xp.float32)
            t=xp.array([example[1] for example in batch],dtype=xp.int32)

            #如果数据集是一维，将他们变为二维，每一列为一个批量
            if x.ndim==1:
                x=x.reshape(-1,1)
            if t.ndim==1:
                t=t.reshape(-1,1)
            self.iteration+=1
            return x,t
        elif len(batch[0])==3:
            x = xp.array([example[0] for example in batch],dtype=xp.float32)
            y = xp.array([example[1] for example in batch],dtype=xp.int32)
            t = xp.array([example[2] for example in batch],dtype=xp.int32)

            # 如果数据集是一维，将他们变为二维，每一列为一个批量
            if x.ndim == 1:
                x = x.reshape(-1, 1)
            if t.ndim == 1:
                t = t.reshape(-1, 1)
            if y.ndim == 1:
                y = y.reshape(-1, 1)
            self.iteration += 1
            return x,y,t
        else:
            raise StopIteration

    def __len__(self):
        '''序列数据的长度等于序列长度除以批次，向上取整'''
        return self.max_iter


import os

import numpy as np
from skystar.sky_dataset.create_dataset import init_dir
def loaddata(file,split=False,split_rate=0.8):
    '''
    split=false retrun: x,t np.ndarray
    split=true retrun: x_train,t_train,x_test,t_test np.ndarray
    '''
    if not os.path.exists(file):
        path = init_dir()
        file=os.path.join(path,file)
    if not os.path.exists(file):
        raise FileNotFoundError
    with np.load(file) as data:
        x = data['x_dataset']
        t = data['t_dataset']
    if split:
        x_train=x[:int(len(x)*split_rate)]
        t_train=t[:int(len(t)*split_rate)]
        x_test=x[int(len(x)*split_rate):]
        t_test=t[int(len(t)*split_rate):]
        return x_train,t_train,x_test,t_test
    return x,t

if __name__ == '__main__':
    x_train,t_train = loaddata('dataset.npz')


import os
import subprocess
import urllib.request
from skystar import no_grad,Variable
from skystar.cuda import get_array_module
from skystar.cuda import as_numpy
from skystar.sky_dataset import data_to_npz
# import matplotlib.pyplot as plt
# import matplotlib.colors as mcolors
import numpy as np
from PIL import Image

from skystar.voc2012 import VOC_COLORMAP  # 颜色条

def make_dir(dirname):
    dir = os.getcwd()
    dir = os.path.join(dir, dirname)
    if not os.path.exists(dir):
        os.makedirs(dir)
    return dir
def pair():
    pass
def _dot_var(v, verbose=False):  # v是Variable实例,该函数实现单个元素的dot转换
    def shape_txt(x):
        label = 'shape['
        for i in range(len(x.shape)):
            _str = str(x.shape[i])
            if i != len(x.shape) - 1:
                label += _str + ' '
            else:
                label += _str
        label += ']'
        return label

    dot_var = '{} [label="{}", color=orange, style=filled]\n'
    if v.name is None:
        name = ' '
    else:
        name = v.name

    if v.data is None:
        name += 'None'
    if verbose and v.data is not None:
        if v.name is not None:
            name += '\n'
        else:
            name = 'X\n'
        name += shape_txt(v) + '\ndtype:' + str(v.dtype)

    return dot_var.format(id(v), name)


def _dot_func(f):  # f是函数，该函数实现输入函数的dot转化
    dot_func = '{} [label="{}", color=lightblue, style=filled, shape=box]\n'
    txt = dot_func.format(id(f), f.__class__.__name__)

    dot_edge = '{} -> {}\n'
    for x in f.inputs:
        txt += dot_edge.format(id(x), id(f))
    for y in f.outputs:
        txt += dot_edge.format(id(f), id(y()))  # y是weakref
    return txt


def get_dot_var(output, verbose=False):
    txt = ''
    funcs = [output.creator]
    txt += _dot_var(output, verbose)

    while funcs:
        func = funcs.pop()
        txt += _dot_func(func)
        for input in func.inputs:
            txt += _dot_var(input, verbose)
            if input.creator is not None:
                if input.creator not in funcs:
                    funcs.append(input.creator)

    return 'digraph g{\n' + txt + '}'


def plot_dot_graph(output, verbose=True, to_file='C:\\Users\\85002\\Desktop\\graph.png'):  # 计算图默认生成在桌面
    dot_graph = get_dot_var(output, verbose)

    # 将dot保存为文件，位置在C:\Uers\85002\.dezer_Cgraph
    tmp_dir = os.path.join(os.path.expanduser('~'), '.dezero_Cgraph')  # 获取文件夹路径/.dezero_Cgraph
    if not os.path.exists(tmp_dir):  # 若文件夹不在，则创建文件夹
        os.makedirs(tmp_dir)
    graph_path = os.path.join(tmp_dir, 'tmp_graph.dot')  # 获取创建文件的路径及名称

    with open(graph_path, 'w') as f:  # 创建文件
        f.write(dot_graph)

    # 外部调用dot命令
    dot_path = r'C:\Program Files\Graphviz\bin\dot.exe'  # dot命令的位置
    extension = os.path.splitext(to_file)[1][1:]  # 获取拓展名
    cmd = f'"{dot_path}" "{graph_path}" -T {extension} -o "{to_file}"'
    ''' cmd=C:\\Program Files\\Graphviz\\bin\\dot.exe C:\\Users\\85002\\.dezero_Cgraph\\tmp_graph.dot -T png -o C:\\Users\\85002\\Desktop\\graph.png'''
    subprocess.run(cmd, shell=True)  # 终端执行cmd命令

    try:
        from IPython import display
        return display.Image(filename=to_file)
    except:
        pass


def sum_to(x, shape):
    """沿轴对元素求和，以输出具有给定形状的数组。
    Parameters：
    x (ndarray): 输入数组。
    shape: 目标形状。
    return：
    ndarray: 具有目标形状的输出数组。
    """
    ndim = len(shape)
    lead = x.ndim - ndim
    lead_axis = tuple(range(lead))

    axis = tuple([i + lead for i, sx in enumerate(shape) if sx == 1])
    out = x.sum(lead_axis + axis, keepdims=True)
    if lead > 0:
        if isinstance(out,Variable):
            out.data = out.data.squeeze(lead_axis)
        else:out = out.squeeze(lead_axis)
    return out


def reshape_sum_backward(gy, x_shape, axis, keepdims):
    """适当地重塑梯度以用于dezero.functions.sum的反向传播。
    Parameters：
    gy (dezero.Variable): 反向传播中从输出获得的梯度变量。
    x_shape (tuple): 在sum函数的前向传播中使用的形状。
    axis (None或int或int的元组): 在sum函数的前向传播中使用的轴。
    keepdims (bool): 在sum函数的前向传播中使用的keepdims参数。
    return：
    dezero.Variable: 适当重塑的梯度变量。
    """
    ndim = len(x_shape)
    tupled_axis = axis
    if axis is None:
        tupled_axis = None
    elif not hasattr(axis, 'len'):
        tupled_axis = (axis,)

    if not (ndim == 0 or tupled_axis is None or keepdims):
        actual_axis = [a if a >= 0 else a + ndim for a in tupled_axis]
        shape = list(gy.shape)
        for a in sorted(actual_axis):
            shape.insert(a, 1)
    else:
        shape = gy.shape

    gy = gy.reshape(shape)
    return gy


def cross_entropy_error(y, t):  # 交叉熵误差，当t为非0-1形式时
    '''
    y[batch,num_classes,H,W],[batch,index,num_classes],[batch,num_classes]
    t[batch,1,H,W],[batch,index],[batch,]
    '''
    xp=get_array_module(y)
    if t.ndim == 4:
        batch,num_classes,H,W = y.shape
        try:
            t=t.squeeze(axis=1)
        except:
            raise ValueError('t.shape must have shape [batch,1,index,num_classes]')
        t=onehot(t,num_classes).transpose(0,3,1,2)
        pt = -xp.sum(t * xp.log(y + 1e-5)) / H / W / batch
    elif t.ndim == 2:
        batch,index,num_classes = y.shape
        t=onehot(t,num_classes)
        pt = -xp.sum(t * xp.log(y + 1e-5)) / index / batch
    elif t.ndim == 1:
        batch, num_classes=y.shape
        t=onehot(t,num_classes)
        pt = -xp.sum(t * xp.log(y + 1e-5)) / batch
    else:
        raise ValueError('t.ndim have to be 2 or 4')
    return pt
    # xp = get_array_module(y)
    # if y.ndim <= 2:
    #     if y.ndim == 1:  # 1把一维数据变为一行的二维数据
    #         y = y.reshape(1, y.size)
    #         t = t.reshape(1, t.size)
    #     if t.size == y.size:  # 当t为0-1形式时提取最大值的索引,变为非0-1形式
    #         t = t.argmax(axis=1)
    #     batch_size = y.shape[0]  # 如果是一维数组的话，这里就会返回不正常值
    #     pt = -xp.sum(xp.log(y[xp.arange(batch_size), t]) + 1e-5) / batch_size
    #     return pt
    # else:
    #     '''这里加入了语义分割模型下，像素级的交叉熵误差'''
    #     if y.ndim == 3:
    #         y = y.reshape(1, y.shape[0], y.shape[1], y.shape[2])
    #     batch_size, num_class, H, W = y.shape
    #     t=t.squeeze()#t的形状可能为1，H，W或者H，W,强制变为H，W
    #     '''将y变为二维数组，与t变为非0-1形式二维数组'''
    #     t = onehot(t, num_class).transpose(2,0,1)
    #     pt = -xp.sum(t * xp.log(y + 1e-5)) / H / W / batch_size
    #     return pt



# def im2col(input_data, filter_h, filter_w, stride=1, pad=0):
#     """
#     将四维图像变为二维矩阵
#     Parameters
#     ----------
#     input_data : 由(数据量, 通道, 高, 宽)的4维数组构成的输入数据
#     filter_h : 滤波器的高
#     filter_w : 滤波器的宽
#     stride : 步幅
#     pad : 填充
#
#     Returns
#     -------
#     col : 2维数组
#     """
#     xp =get_array_module(input_data)
#     N, C, H, W = input_data.shape
#     out_h = (H + 2*pad - filter_h)//stride + 1#输出窗口高
#     out_w = (W + 2*pad - filter_w)//stride + 1#输出窗口宽
#
#     img = xp.pad(input_data, [(0,0), (0,0), (pad, pad), (pad, pad)], 'constant')
#     col = xp.zeros((N, C, filter_h, filter_w, out_h, out_w))
#     '''经过下面的循环，将col变成（N,C,filter_h,filter_w,out_h,out_w）'''
#     for y in range(filter_h):
#         y_max = y + stride*out_h
#         for x in range(filter_w):
#             x_max = x + stride*out_w
#             col[:, :, y, x, :, :] = img[:, :, y:y_max:stride, x:x_max:stride]
#     '''数据的展开是按照0，1，2，3....的顺序依次展开的，注意核作用的区域大小为（filter_h,filter_w）
#     所以为了正确展开，对数据的轴重新排列（N,out_h,out_w,C,filter_h,filter_w）'''
#     col = col.transpose(0, 4, 5, 1, 2, 3).reshape(N*out_h*out_w, -1)
#     return col#(N*oh*ow,c*fh*fw)
def im2col(input_data, filter_h, filter_w, stride=1, pad=0):  # 一种更加高效的im2col方法
    xp = get_array_module(input_data)
    N, C, H, W = input_data.shape
    out_h = (H + 2 * pad - filter_h) // stride + 1
    out_w = (W + 2 * pad - filter_w) // stride + 1
    # 添加填充
    img = xp.pad(input_data, ((0, 0), (0, 0), (pad, pad), (pad, pad)), mode='constant')
    # 生成一个索引数组
    i0 = xp.repeat(xp.arange(filter_h), filter_w)
    i1 = xp.tile(xp.arange(filter_w), filter_h)
    j = xp.repeat(xp.arange(out_h), out_w)
    k = xp.tile(xp.arange(out_w), out_h)
    i = i0[:, None] + j[None, :] * stride
    j = i1[:, None] + k[None, :] * stride
    col = img[:, :, i, j].transpose(0, 3, 1, 2).reshape(N * out_h * out_w, -1)
    return col


# def col2im(col, input_shape, filter_h, filter_w, stride=1, pad=0):
#     """
#     将二维矩阵变为四维图像
#     Parameters
#     ----------
#     col :
#     input_shape : 输入数据的形状（例：(10, 1, 28, 28)）
#     filter_h :窗口高度
#     filter_w:窗口宽度
#     stride:步长
#     pad:填充
#
#     Returns
#     -------
#     """
#     xp =get_array_module(col)
#     N, C, H, W = input_shape
#     out_h = (H + 2*pad - filter_h)//stride + 1
#     out_w = (W + 2*pad - filter_w)//stride + 1
#     col = col.reshape(N, out_h, out_w, C, filter_h, filter_w).transpose(0, 3, 4, 5, 1, 2)
#
#     img = xp.zeros((N, C, H + 2*pad + stride - 1, W + 2*pad + stride - 1))
#     for y in range(filter_h):
#         y_max = y + stride*out_h
#         for x in range(filter_w):
#             x_max = x + stride*out_w
#             img[:, :, y:y_max:stride, x:x_max:stride] += col[:, :, y, x, :, :]
#
#     return img[:, :, pad:H + pad, pad:W + pad]#(N,C,H,W)
def col2im(col, input_shape, filter_h, filter_w, stride=1, pad=0):
    xp = get_array_module(col)
    N, C, H, W = input_shape
    out_h = (H + 2 * pad - filter_h) // stride + 1
    out_w = (W + 2 * pad - filter_w) // stride + 1
    # 重塑并转置col
    col = col.reshape(N, out_h, out_w, C, filter_h, filter_w).transpose(0, 3, 4, 5, 1, 2)
    # 初始化图像
    img = xp.zeros((N, C, H + 2 * pad, W + 2 * pad))
    # 计算每个窗口的填充区域并进行累加
    for y in range(filter_h):
        for x in range(filter_w):
            img[:, :, y:y + out_h * stride:stride, x:x + out_w * stride:stride] += col[:, :, y, x, :, :]
    return img[:, :, pad:H + pad, pad:W + pad]  # 返回裁剪后的图像


def pair(x):
    if isinstance(x, int):
        return (x, x)
    elif isinstance(x, tuple):
        assert len(x) == 2
        return x
    else:
        raise ValueError


def show_progress(block_num, block_size, total_size):
    bar_template = "\r[{}] {:.2f}%"

    downloaded = block_num * block_size
    p = downloaded / total_size * 100
    i = int(downloaded / total_size * 30)
    if p >= 100.0: p = 100.0
    if i >= 30: i = 30
    bar = "#" * i + "." * (30 - i)
    print(bar_template.format(bar, p), end='')


# cache_dir = os.path.join(os.path.expanduser('~'), '.dezero')#'C:\\Users\\85002\\.dezero'
cache_dir = 'D:\\Programing\\pythonProject\\Dezero\\train_image'


def get_file(url, file_name=None):
    """Download a file from the `url` if it is not in the cache.

    The file at the `url` is downloaded to the `~/.dezero`.

    Args:
        url (str): URL of the file.
        file_name (str): Name of the file. It `None` is specified the original
            file name is used.

    Returns:
        str: Absolute path to the saved file.
    """
    if file_name is None:
        file_name = url[url.rfind('/') + 1:]
    file_path = os.path.join(cache_dir, file_name)

    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)

    if os.path.exists(file_path):
        return file_path

    print("Downloading: " + file_name)
    try:
        urllib.request.urlretrieve(url, file_path, show_progress)
    except (Exception, KeyboardInterrupt) as e:
        if os.path.exists(file_path):
            os.remove(file_path)
        raise
    print(" Done")

    return file_path


def transconv_pad(x, stride=1, kernel_size=3, pad=0):
    '''对x实现边缘填充和中间填充，扩展图像的尺寸'''
    xp = get_array_module(x)
    middle_pad = stride - 1
    edge_pad = kernel_size - 1 - pad  # 这里说明pad影响了边缘填充
    N, C, h, w = x.shape

    # 计算中间填充后图像尺寸
    new_h = h + (h - 1) * middle_pad
    new_w = w + (w - 1) * middle_pad

    mid_demox = xp.zeros((N, C, new_h, new_w))

    for i in range(h):
        for j in range(w):
            mid_demox[:, :, i * (1 + middle_pad), j * (1 + middle_pad)] = x[:, :, i, j]
    # 使用edge_pad在图像边缘添加填充
    x = xp.pad(mid_demox, [(0, 0), (0, 0), (edge_pad, edge_pad), (edge_pad, edge_pad)], mode='constant')
    return x

def back_transcov_pad(x, stride=1, kernel_size=3, pad=0):
    '''从反卷积结果中还原原始图像，用于上采样层的反向传播'''
    xp = get_array_module(x)
    edge_pad = kernel_size - 1 - pad
    middle_pad = stride - 1
    N, C, h, w = x.shape
    # 去除边缘填充
    x = x[:, :, edge_pad:h - edge_pad, edge_pad:w - edge_pad]
    h, w = x.shape[2], x.shape[3]
    # 计算去除中间填充后的尺寸
    new_h = int((h + middle_pad) // (1 + middle_pad))
    new_w = int((w + middle_pad) // (1 + middle_pad))
    out = xp.zeros((N, C, new_h, new_w))

    for i in range(new_h):
        for j in range(new_w):
            out[:, :, i, j] = x[:, :, i * stride, j * stride]

    return out

def bilinear_kernel(in_channels, out_channels, kernel_size, xp=np):
    '''创建一个双线性内插值的核，用于初始化上采样层的权重'''
    factor = (kernel_size + 1) // 2
    if kernel_size % 2 == 1:
        center = factor - 1
    else:
        center = factor - 0.5
    og = (xp.arange(kernel_size).reshape(-1, 1),
          xp.arange(kernel_size).reshape(1, -1))
    filt = (1 - xp.abs(og[0] - center) / factor) * \
           (1 - xp.abs(og[1] - center) / factor)
    weight = xp.zeros((in_channels, out_channels, kernel_size, kernel_size))
    weight[:] = filt
    return xp.array(weight)


def onehot(x, num_classes):
    '''
    此函数用于把标签t变为onehot形式,默认轴为最后一轴，本函数不改变数据轴
    支持输入：[batch,index],[batch,H,W]
    输出：[batch,index,num_classes],[batch,H,W,num_classes]
    '''
    xp = get_array_module(x)
    x=x.astype(xp.int32)
    return xp.eye(num_classes)[x]
def word_to_idx(x, num_classes):
    xp=get_array_module(x)
    batch,seq_len=x.shape
    x=x.reshape(-1)
    rex = xp.zeros((x.size, num_classes))
    rex[xp.arange(seq_len*batch), x] = 1
    rex = rex.reshape(batch, seq_len, num_classes)
    return rex

def accuracy(y, t):
    '''
    y:Variable
    t:ndarray
    '''
    with no_grad():
        xp = get_array_module(y)
        if t.ndim > 3:
            xp = get_array_module(y)
            y = y.data.argmax(axis=1)
            num_classes, H, W = y.shape
            y = y.reshape(num_classes, 1, H, W)
            acc = xp.sum(y == t) / y.size
            return acc
        else:
            if t.ndim == 2:
                t = xp.argmax(t, axis=1)
            y = xp.argmax(y.data, axis=1)
            sum = t.size
            return xp.sum(y == t) / sum


def mean_IoU(y, t):
    '''
    y:Variable 预测值 四维
    t:ndarray 真实标签 四维
    '''
    xp = get_array_module(y)
    if y.data.size != t.size:
        y = y.data.argmax(axis=1)  # 预测标签
    num_classes, H, W = y.shape
    y = y.reshape(num_classes, 1, H, W)
    unique = xp.unique(t)
    iou_list = []
    for i in unique:
        true_positive = xp.sum((t == i) & (y == i))  # 正确分类的总像素
        false_positive = xp.sum((t != i) & (y == i))  # 错误分类的总像素
        false_negative = xp.sum((t == i) & (y != i))  # 被漏掉的总像素
        denominator = true_positive + false_positive + false_negative
        if denominator == 0:  # 避免除以零
            iou = 0.0
        else:
            iou = true_positive / denominator
        iou_list.append(iou)
    mean_iou = xp.mean(xp.array(iou_list))  # 计算平均IoU
    mean_iou = float(mean_iou)
    return mean_iou
def dataset_to_npz(data_txt_name, outfile_name):
    '''输入txt文件，返回数据集，数据集的字典为x_dataset,t_dataset'''
    return data_to_npz(data_txt_name, outfile_name)

def data_norm(x, _range=(0, 1),dtype=np.float32):
    '''
    对单幅图像进行归一化
    :param x: 图像的矩阵，shape：num,C,H,W
    :param _range: 归一化范围
    :return: 返回归一化结果
    '''
    range_min = _range[0]
    range_max = _range[1]
    _max = np.max(x)
    _min = np.min(x)
    k = (range_max - range_min) / (_max - _min + 1e-5)
    x = k * (x - _min)
    x=x.astype(dtype)
    return x


# def datatofeature(data,mode='feature',pad=1):
#     '''图像预处理'''
#     data = skystar.cuda.as_numpy(data)
#     if mode == 'feature' or mode == 'weight':
#         data = data_norm(data, _range=(0, 255))  # 把数据映射到0~255
#     data = data.astype(np.float32)
#     data = data.transpose(0, 2, 3, 1)
#
#     '''额外添加图片，满足图像x*x排列'''
#     num, H, W, C = data.shape
#     x = int_square_root(num)  # x*x
#     if x * x > num:
#         add_num = x * x - num
#         new_data = np.zeros((add_num, H, W, C), dtype=np.float32)
#         if mode == 'feature' or 'weight':
#             new_data += 255  # 生成白图
#         elif mode == 'label':
#             new_data += np.max(data)
#         data = np.concatenate((data, new_data), axis=0)
#     full_image = splicing(data, x, pad=pad, padnum=255)  # 图像填充和拼接
#     return full_image
# def int_square_root(x):
#     '''确定x的整数平方根，若无整数平方根，x+1直到能取整数平方根为之'''
#     sqrt = np.sqrt(x)
#     y = asif_int(sqrt)
#     while not y:
#         x += 1
#         sqrt = np.sqrt(x)
#         y = asif_int(sqrt)
#     return int(sqrt)
# def asif_int(x):
#     '''判断数是否为整数'''
#     x = ((x - int(x)) == 0)
#     return x
# def splicing(data, P_num, pad=1, padnum=255):
#     '''该函数用于将四维的多个图片变为二维矩阵，即将图像二维排列,P_num:排列数'''
#     num, H, W, C = data.shape  # 原图大小
#     data = data.reshape(P_num, P_num, H, W, C)
#     # 按行和列拼接数据，形成完整的图像
#     rows = []
#     for i in range(P_num):
#         row = np.concatenate(data[i, :, :, :, :],
#                              axis=1)  # 拼接行data[i, :, :, :, :]取出 i行(x, H_pad, W_pad, C)数据,拼接后的数据降维（H，W，C），可以理解为axis=1在W方向上拼接
#         rows.append(row)
#     full_image = np.concatenate(rows, axis=0)
#     # 给不同图像之间添加分界线用以区别
#     H_full, W_full, C = full_image.shape
#     re_img = (np.zeros((H_full + P_num - 1, W_full + P_num - 1, C)) + padnum)
#     for i in range(P_num):
#         x = i * (H + pad)
#         for j in range(P_num):
#             y = j * (W + pad)
#             re_img[x:x + H, y:y + W, :] = full_image[i * H:i * H + H, j * W:j * W + W, :]
#     return re_img
# def X_mode_show(feature, label,alpha=1):
#     H, W, C = feature.shape
#     if C == 1:  # 通道为1显示灰度图
#         plt.imshow(feature, cmap='gray')
#     else:  # 真彩色
#         re_img = np.zeros((H, W, 3))
#         re_img[:, :, 0] = feature[:, :, 0]
#         re_img[:, :, 1] = feature[:, :, 1]
#         if C == 2:
#             re_img[:, :, 2] += 255
#         else:
#             re_img[:, :, 2] = feature[:, :, 2]
#         re_img = np.uint8(re_img)
#         plt.imshow(re_img,alpha=alpha)
#     plt.axis('off')
#     plt.title('Feature' + label)
#     plt.show()
#
# def T_mode_show(feature, label):
#     feature = np.squeeze(feature)
#     unique_labels = np.unique(feature)
#     cmap = voc_colormap2cmap()  # 使用自定义的颜色条映射
#     # 创建标签与颜色之间的映射
#     norm = mcolors.BoundaryNorm(boundaries=np.arange(len(unique_labels) + 1) - 0.5, ncolors=len(unique_labels))
#     plt.imshow(feature, cmap=cmap, norm=norm,alpha=1)
#     plt.axis('off')
#     plt.title('Label' + label)
#     # 显示图像
#     plt.show()
# def Weight_mode_show(feature, label):
#     H, W, C = feature.shape
#     if C == 1:  # 通道为1显示灰度图
#         plt.imshow(feature, cmap='gray')
#     else:
#         re_img = np.zeros((H, W, 3))
#         re_img[:, :, 0] = feature[:, :, 0]
#         re_img[:, :, 1] = feature[:, :, 1]
#         if C == 2:
#             re_img[:, :, 2] += 255
#         else:
#             re_img[:, :, 2] = feature[:, :, 2]
#         re_img = re_img[:, :, 0] * 0.2989 + re_img[:, :, 1] * 0.5870 + re_img[:, :, 2] * 0.1140
#         re_img = np.uint8(re_img)
#         plt.imshow(re_img, cmap='gray')
#     plt.axis('off')
#     plt.title('Feature' + label)
#     plt.show()
#     return
# def images_show(data, pad=1, mode='feature', label=None):
#     '''
#     :param data: ndarray shape: num,C,H,W
#     :param pad: 中间填充数量，用于区分图像，默认为1
#     :param mode: 模式，默认为’X‘  可选X T
#     :param label: 图名
#     :return: 显示平铺的图形
#     '''
#     '''图像预处理'''
#     full_image=datatofeature(data,mode=mode,pad=pad)
#     '''图像显示'''
#     if label is None:
#         label = ''
#     else:
#         label = ':' + label
#     if mode == 'feature':
#         X_mode_show(full_image, label)
#     elif mode == 'label':
#         T_mode_show(full_image, label)
#     elif mode == 'weight':
#         Weight_mode_show(full_image, label)
#     else:
#         print(f'not support mode:{label}')

# def init_graph():
#     figure,axes= plt.subplots(nrows=1, ncols=3)
#     for ax in axes.flat:  # 对于多行，遍历所有子图并关闭轴
#         ax.axis('off')
#     return figure,axes
# def subplots_show(axes, x, t, predict, t_alpha=1):
#     feature = datatofeature(x, mode='feature')
#     t_feature=datatofeature(t,mode='label')
#     predict_feature=datatofeature(predict,mode='label')
#     fh,fw,fc=feature.shape
#     re_img = np.zeros((fh, fw, 3))
#     if fc>=3:
#         re_img[:,:,0:3]=feature[:,:,0:3]
#         re_img = np.uint8(re_img)
#         axes[0].imshow(re_img)
#
#     else:
#         print('真彩图的通道数少于3，无法正常显示')
#         raise
#     t_feature = np.squeeze(t_feature)
#     predict_feature = np.squeeze(predict_feature)
#
#     unique_labels = np.unique(t_feature)
#     cmap = voc_colormap2cmap()  # 使用自定义的颜色条映射
#     # 创建标签与颜色之间的映射
#     norm = mcolors.BoundaryNorm(boundaries=np.arange(len(unique_labels) + 1) - 0.5, ncolors=len(unique_labels))
#     axes[1].imshow(t_feature, cmap=cmap, norm=norm)
#     axes[2].imshow(predict_feature, cmap=cmap, norm=norm,alpha=t_alpha)

# def save_figure(figure, label=None):
#     dir = os.getcwd()
#     dir = os.path.join(dir, 'figures')
#     if not os.path.exists(dir):
#         os.makedirs(dir)
#     filename = os.path.join(dir, label)
#     figure.savefig(filename)
#     print(f'Figture creted!path:{filename}')

def write_text(list,filename):
    dir = os.getcwd()
    dir = os.path.join(dir, 'results')
    if not os.path.exists(dir):
        os.makedirs(dir)
    filename=os.path.join(dir, filename)
    with open(filename,'w',encoding='utf-8') as f:
        for item in list:
            f.write(str(item)+'\n')
    print(f'list.txt created path:{filename}')
def read_text(filename):
    dir = os.getcwd()
    dir = os.path.join(dir, 'results')
    if not os.path.exists(dir):
        os.makedirs(dir)
    filename=os.path.join(dir, filename)
    with open(filename,'r',encoding='utf-8') as f:
        text_list = [float(line.strip()) for line in f if line.strip()]
    return text_list
# def voc_colormap2cmap():
#     """将自定义的颜色条转换为 matplotlib 可识别的 ListedColormap"""
#     return mcolors.ListedColormap(np.array(VOC_COLORMAP) / 255.0)
def To_PIL(data_array):
    if data_array.shape[0]==3:
        data_array=data_array.transpose((1,2,0))
    img=Image.fromarray(data_array)
    return img
def label_to_rgb(t):
    '''根据颜色条将label映射到rgb图像'''
    H, W = t.shape
    t=t.reshape(-1)
    rgb=[VOC_COLORMAP[i] for i in t ]
    rgb=np.array(rgb,dtype=np.uint8)
    rgb=rgb.reshape(H,W,3)
    return rgb
def image_pad(im):
    '''给图像填充1像素的黑边'''
    data=np.array(im)
    data=np.pad(data,((1,1),(1,1),(0,0)), 'constant',constant_values=0)
    return To_PIL(data)
def cal_premutation(num):
    '''输入图像的数量，计算最佳排列，如12->3*4,13->4*4'''
    sqrt=np.sqrt(num)
    int_sqrt=int(sqrt)
    if num==int_sqrt*int_sqrt:
        return (int_sqrt,int_sqrt)
    elif (int_sqrt*(int_sqrt+1)>=num) and (num>int_sqrt*int_sqrt):
        return (int_sqrt,int_sqrt+1)
    else:
        return (int_sqrt+1,int_sqrt+1)

def splicing(data):
    '''把四维的array数据转化为多个图像并拼接'''
    # 把四维data转化为多个三维datas
    data = as_numpy(data)
    num, C, H, W = data.shape
    if C==3:
        datas=[data_norm(data[i],(0,255),np.uint8) for i in range(num)]#对数据进行归一化
    elif C==1:
        datas=[label_to_rgb(data[i][0]) for i in range(num)]
    else:
        raise
    im_list=[To_PIL(data) for data in datas]
    im_list=[image_pad(im) for im in im_list]

    # 单幅图像尺寸
    width, height = im_list[0].size

    # 创建空图
    h_num,w_num=cal_premutation(len(im_list))
    result=Image.new(im_list[0].mode, (width*w_num, height*h_num))

    # 图像拼接
    for i in range(h_num):
        for j in range(w_num):
            index=i*w_num+j
            result.paste(im_list[index],(width*j,height*i))
            if (index+1)==len(im_list):
                break
    return result
def Create_ImgForSeg(datas):
    '''接受一个元组，为元组里的数据创建对比图，期望元组（x,t,predict），数据的shape（N，C，H，W）'''
    datas=[as_numpy(data) for data in datas]
    w_num=len(datas)
    h_num=len(datas[0])
    imgs=[]
    for i in range(w_num):
        data=datas[i]
        N, C, H, W = data.shape
        if C==3:
            imgdatas=[data_norm(data[i],(0,255),np.uint8) for i in range(N)]
        elif C==1:
            imgdatas=[label_to_rgb(data[i][0]) for i in range(N)]
        else:
            raise ValueError('数据的通道数不符合要求，要求通道为3或1')
        imglist = [To_PIL(data) for data in imgdatas]
        imglist = [image_pad(im) for im in imglist]
        imgs.append(imglist)
    width, height = imgs[0][0].size
    result=Image.new(imgs[0][0].mode,(width*w_num,height*h_num))
    for i in range(w_num):
        for j in range(h_num):
            imglist=imgs[i]
            result.paste(imglist[j],(width*i,height*j))
    return result


def save_img(im,name='demo.jpg'):
    dir = make_dir('figures')
    if '.jpg' not in name:
        name+='.jpg'
    # 保存图片
    im.save(os.path.join(dir,name))


import numpy as np
import skystar.sky_dataset
import os
class Dataset:
    def __init__(self,training=True,transform=None,target_transform=None):
        self.training=training
        self.transform=transform
        self.target_transform=target_transform
        if self.transform is None:
            self.transform=lambda x:x#设置临时函数，将数据按照原样输出
        if self.target_transform is None:
            self.target_transform=lambda x:x
        self.data=None
        self.label=None
        self.prepare()
    def __getitem__(self, index):
        if isinstance(index, (int, np.integer)):  # 处理单个标量索引
            if self.label is None:
                return self.transform(self.data[index]), None
            else:
                return self.transform(self.data[index]), self.target_transform(self.label[index])
        elif isinstance(index, slice):  # 处理切片索引
            if self.label is None:
                return self.transform(self.data[index]), None
            else:
                return (self.transform(self.data[index]),
                        self.target_transform(self.label[index]))
        else:
            raise TypeError(f'Invalid index type: {type(index)}')

    def __len__(self):
        return len(self.data)

    def prepare(self):
        pass

class BigData(Dataset):
    def __getitem__(self, index):
        x=np.load(f'data/{index}.npy')
        t=np.load(f'label/{index}.npy')

        return x,t
    def __len__(self):
        return 1000000

class MathImg_data(Dataset):
    def __init__(self,normalize=True,flatten=True,one_hot_label=False,training=True,transform=None,target_transform=None):
        self.normalize=normalize
        self.flatten=flatten
        self.one_hot_label=one_hot_label
        super().__init__(training,transform,target_transform)
    def prepare(self):
        (x_train, t_train), (x_test, t_test) = load_mnist(normalize=self.normalize, flatten=self.flatten, one_hot_label=self.one_hot_label)
        if self.training:
            self.data=x_train
            self.label=t_train
        else:
            self.data=x_test
            self.label=t_test

class Stl_10(Dataset):
    def __init__(self,training=True,transform=None,target_transform=None):
        super().__init__(training,transform,target_transform)

    def prepare(self):
        path=os.path.abspath(__file__)
        path=os.path.dirname(path)
        if self.training:
            X_path=os.path.join(path,'STL10\\stl10_binary','train_X.bin')
            Y_path=os.path.join(path,'STL10\\stl10_binary','train_y.bin')
            x_train = read_all_images(X_path)
            self.data = x_train.transpose(0, 3, 1, 2)
            self.label = read_labels(Y_path)
            self.label=self.label-1#该数据的标签从1开始
        else:
            X_path=os.path.join(path,'STL10\\stl10_binary','test_X.bin')
            Y_path=os.path.join(path,'STL10\\stl10_binary','test_y.bin')
            x_test=read_all_images(X_path)
            self.data=x_test.transpose(0,3,1,2)
            self.label=read_labels(Y_path)
            self.label=self.label-1#该数据的标签从1开始

class Tiny_image(Dataset):
    def __init__(self,training=True,transform=None,target_transform=None):
        super().__init__(training,transform,target_transform)

    def prepare(self):
        x_train, t_train, x_test, t_test=load_data()
        if self.training:
            self.data,self.label=x_train,t_train
        else:
            self.data,self.label=x_test,t_test

class selfdata_starsky(Dataset):
    def __init__(self,filename,training=True,transform=None,target_transform=None,split=True):
        '''默认将数据集分割为训练集与验证集'''
        self.filename=filename
        self.split=split
        super().__init__(training,transform,target_transform)

    def prepare(self):
        if self.split:
            x_train, t_train, x_test, t_test = skystar.sky_dataset.loaddata(self.filename,split=self.split)
            if self.training:
                self.data, self.label = x_train, t_train
            else:
                self.data, self.label = x_test, t_test
        else:
            x, t = skystar.sky_dataset.loaddata(self.filename,split=self.split)
            return x,t

class Sindata(Dataset):
    '''三角函数测试数据集，用于时序模型测试'''
    def __init__(self,training=True,size=1000):
        super().__init__(training)
        self.size=size
        self.timestep=np.arange(0,size,1)
        self.Seqdata=np.sin(np.linspace(0,4*np.pi,size))+np.random.normal(-0.025,0.025,size)
        self.ForPredictData=np.cos(np.linspace(0,4*np.pi,size))

    def prepare(self):
        self.data=self.Seqdata[0:self.size-1]
        self.label=self.Seqdata[1:self.size]

class TranslationDataset(Dataset):
    '''用于transformer的翻译数据集，输出三个数据，enc_inputs（编码器输入），dec_inputs（解码器输入），dec_outputs（解码器输出）'''
    def __init__(self,sentences, src_vocab, tgt_vocab,training=True):
        super().__init__(training)
        self.src_vocab = src_vocab
        self.tgt_vocab = tgt_vocab
        self.sentences = sentences

        self.init_params()
        self.enc_inputs,self.dec_inputs,self.dec_outputs=self.make_data()
    def __getitem__(self, index):
        if isinstance(index, (int, np.integer)):  # 处理单个标量索引
                return self.enc_inputs[index], self.dec_inputs[index],self.dec_outputs[index]
        elif isinstance(index, slice):  # 处理切片索引
                return self.enc_inputs[index],self.dec_inputs[index],self.dec_outputs[index]
        else:
            raise TypeError(f'Invalid index type: {type(index)}')
    def __len__(self):
        return len(self.enc_inputs)
    def init_params(self):
        self.src_vocab_size = len(self.src_vocab)#源字典大小
        self.tgt_vocab_size = len(self.tgt_vocab)#目标字典大小
        self.src_len = len(self.sentences[0][0].split(" "))#源句子长度
        self.tgt_len = len(self.sentences[0][1].split(" "))#目标句子长度
    def make_data(self):
        enc_inputs, dec_inputs, dec_outputs = [], [], []
        for i in range(len(self.sentences)):
            enc_input = [[self.src_vocab[n] for n in self.sentences[i][0].split()]]
            dec_input = [[self.tgt_vocab[n] for n in self.sentences[i][1].split()]]
            dec_output = [[self.tgt_vocab[n] for n in self.sentences[i][2].split()]]
            enc_inputs.extend(enc_input)
            dec_inputs.extend(dec_input)
            dec_outputs.extend(dec_output)
        return np.float32(enc_inputs), np.float32(dec_inputs), np.float32(dec_outputs)


import weakref
import numpy as np
import contextlib
import skystar
try:
    import cupy
    array_types=(np.ndarray,cupy.ndarray)
    cupy.set_printoptions(precision=4)
except ImportError:
    array_types=(np.ndarray)
np.set_printoptions(precision=4)
TrainingMode = True
def Set_TrainingMode(mode: bool):
    if not isinstance(mode, bool):
        raise ValueError("Mode must be a boolean value (True or False).")
    global TrainingMode
    TrainingMode = mode
    if TrainingMode:
        print('========Setting mode to Training========')
    else:
        print('========Setting mode to Testing========')
def Get_TrainingMode():
    return TrainingMode

class Variable:
    '''--------
    data：输入数据，仅支持ndarray结构
    name:数据名称
    creator:创造输出的函数，该函数具有input和output，其中input==data
    grad：函数节点处的梯度
    generation:数据的辈分，用于反向传播时决定传播的顺序
    '''
    def __init__(self,data,name=None,):
        if data is not None:#如果数据不是ndarray则报错
            if not isinstance(data, array_types):
                raise TypeError('{} is not supposed'.format(type(data)))
        self.data=data
        self.name=name
        self.grad=None
        self.creator=None
        self.generation=0
    '''为每个数据连接函数并辈分'''
    def set_creator(self,func):
        self.creator=func
        self.generation=func.generation+1
    def cleangrad(self):#用以删除传递中间过程产生数据的梯度，释放内存
        self.grad=None
    def unchain(self):#删除数据与函数的链接
        self.creator=None
    def unchain_backward(self):#用于截断时间序列模型的连接（必须使用，不然模型的计算图无限堆叠）
        if self.creator is not None:
            funcs=[self.creator]
            while funcs:
                f=funcs.pop()
                for x in f.inputs:
                    if x.creator is not None:
                        funcs.append(x.creator)
                        x.unchain()
    '''用于把数据用图像展示，采用了matplotlib的接口,暂时取消使用'''
    def img_show(self):
        img=skystar.utils.splicing(self.data)
        img.show()
    def data_to_img_save(self,out_filename=None):
        if out_filename is None and self.name is None:
            out_filename='NoneName'
        img=skystar.utils.splicing(self.data)
        skystar.utils.save_img(img,out_filename)
    def to_cpu(self):
        if self.data is not None:
            self.data=skystar.cuda.as_numpy(self.data)

    def to_gpu(self):
        if self.data is not None:
            self.data=skystar.cuda.as_cupy(self.data)

    def to_float32(self,dtype=None):
        xp=skystar.cuda.get_array_module(self.data)
        dtype=xp.float32
        self.data=self.data.astype(dtype)

    def backward(self,retain_grad=False,create_graph=False):
        '''
        循环结构完成反向传播
        retain_grad:是否保留反向过程中产生的grad
        create_grad:是否启用高阶求导
        '''
        if Config.enable_backprop==False:#反向传播的开关，默认为Ture
            return None
        if self.grad is None:#初始化为Variable，便于计算导数的导数
            xp=skystar.cuda.get_array_module(self.data)
            self.grad=Variable(xp.ones_like(self.data,dtype=xp.float32))#这里把grad变成了
        funcs=[self.creator]
        while funcs:
            f=funcs.pop()
            dys = [output().grad for output in f.outputs]#list，元素为Variable
            with config_using('enable_backprop',create_graph):
                '''如果create_graph为False,在下面的计算中enable_backprop设置为False，由于运算符重载，那么下面的dxs = f.backward(*dys)将会
                调用Function.__call__方法，Function会根据enable_backprop禁止grad（Variable）链接函数f，由此梯度的反向传播无法进行。'''
                dxs = f.backward(*dys)  # 多个值返回元组，一个值返回单个元素，元素为Variable
                if not isinstance(dxs, tuple):
                    dxs = (dxs,)  # 单个值转化为元组，元素为Variable

                '''对每个input设置导数,如果输入两个数据是同一个数据，那么该数据的导数要叠加。这里
                同时会产生一个问题，如果对一个相同的变量反复反向传播，那么它的梯度会叠加，因此需要在
                每次传播后使用self.cleangrad()初始化梯度'''
                for x, dx in zip(f.inputs, dxs):  # x:Variable,dx:Variable
                    # if isinstance(dx, Variable):
                    #     dx.to_float32()
                    # else:
                    #     xp=skystar.cuda.get_array_module(dx)
                    #     dx=dx.astype(xp.float32)
                    if x.grad is None:
                        x.grad = dx
                    else:
                        x.grad = x.grad + dx
                    if x.creator is not None:
                        if x.creator not in funcs:  # 这里避免相同函数的backward被调用两次
                            funcs.append(x.creator)
                            funcs.sort(key=lambda x: x.generation)  # key为函数，指定取待排序元素的哪一项进行排序

                    if not retain_grad:  # 这里删除了中间过程产生的实例变量的grad，释放内存，由用户输入的数据grad的保存了下来
                        for y in f.outputs:
                            y().grad = None
    def reshape(self, *shape):
        if len(shape) == 1 and isinstance(shape[0], (tuple, list)):
            shape = shape[0]
        return reshape(self, shape)

    def transpose(self, *axes):
        if len(axes) == 0:
            axes = None
        elif len(axes) == 1:
            if isinstance(axes[0], (tuple, list) or axes[0] is None):
                axes = axes[0]
        return transpose(self, axes)

    def sum(self,axis=None,keepdims=False):
        return sum(self,axis,keepdims)
    '''静态方法，可以在调用时不加括号，比如x=Variable(np.array),可以直接实现x.shape查看数据形状'''
    @property
    def shape(self):
        return self.data.shape
    @property
    def ndim(self):
        return self.data.ndim
    @property
    def size(self):
        return self.data.size
    @property
    def dtype(self):
        return self.data.dtype
    @property
    def T(self):
        return transpose(self)

    __array_priority__=200#设置运算优先级，防止ndarray与Variable运算时，调用的是numpy的运算符
    def __len__(self):#实现Variable实例直接调用len函数
        return len(self.data)

    def __repr__(self):#改变Variable实例的打印方式
        if self.data is None:
            return 'Variable(None)'
        p=str(self.data).replace('\n','\n'+' '*9)
        return 'Variable('+p+')'

    def __getitem__(self, key):
        # 支持切片读取
        return self.data[key]

    def __setitem__(self, key, value):
        # 支持切片赋值
        self.data[key] = value
    '''重载运算符，实现Variable实例 加减乘除等方式,函数setup_variable也实现了该功能'''

class Parameter(Variable):#创建子集Parameter，用来存储模型的参数
    pass



class Function:
    '''----------
    inputs：#可以是多个数据，使用加减乘除时必须包含一个Variable实例
    self.inputs：#list，元素为Variable
    self.outputs：#list,元素为Variable实例,使用了弱引用，调用的话需要写为：元素（）
    generation：函数的辈分
    '''
    def __call__(self, *inputs):#添加*号，可以同时输入数据，将多个数据变为元组tuple
        inputs=[as_variable(x) for x in inputs]#将输入变为list,元素为Variable实例
        xs=[x.data for x in inputs]#list,元素为ndarray
        ys=self.forward(*xs)#使用*号解包，传向forward的inputs为元组或单个元素，元素为ndarray
        if not isinstance(ys,tuple):#如果不是元组进行处理
            ys=(ys,)
        outputs=[Variable(as_scalar(y)) for y in ys]#list，元素为Variable

        if Config.enable_backprop:
            self.generation = max([x.generation for x in inputs])
            for output in outputs:
                output.set_creator(self)

        self.inputs=inputs#list,元素为Variable实例
        self.outputs=[weakref.ref(output) for output in outputs]#这里添加了弱引用，减少内存占用，如果要调用列表元素x的值则需要使用x()

        if len(outputs)>1:
            return outputs
        else:
            return outputs[0]#Variable

    def forward(self,x):
        raise NotImplementedError()

    def backward(self,dout):
        raise NotImplementedError()

# =============================================================================
'''四则运算和常用运算'''
# =============================================================================
class Exp(Function):
    def forward(self,x):
        xp=skystar.cuda.get_array_module(x)
        y=xp.exp(x)
        return y
    def backward(self,dout):
        x=self.inputs[0]
        xp=skystar.cuda.get_array_module(dout)
        dout=xp.exp(x)*dout
        return dout
class Square(Function):
    def forward(self,x):
        y=x**2
        return y
    def backward(self,dout):
        dout=2*self.inputs[0]*dout
        return dout
class Add(Function):#加法,实现了广播功能
    def forward(self,x0,x1):
        self.x0_shape,self.x1_shape=x0.shape,x1.shape
        x1=as_scalar(x1)
        return x0+x1
    def backward(self,dout):
        dout0,dout1=dout,dout
        if self.x0_shape != self.x1_shape:
            dout0=skystar.utils.sum_to(dout0,self.x0_shape)
            dout1=skystar.utils.sum_to(dout1,self.x1_shape)
        return dout0,dout1
class Mul(Function):#乘法
    def forward(self,x0,x1):
        self.x0_shape,self.x1_shape=x0.shape,x1.shape
        x1=as_scalar(x1)
        return x0*x1
    def backward(self,dout):
        x0,x1=self.inputs
        if self.x0_shape != self.x1_shape:
            dout0=skystar.utils.sum_to(dout*x1,self.x0_shape)
            dout1=skystar.utils.sum_to(dout*x0,self.x1_shape)
        else:
            dout0 = dout * x1
            dout1 = dout * x0
        return dout0,dout1
class Neg(Function):#负数
    def forward(self,x):
        return -x
    def backward(self,dout):
        return -1*dout
class Sub(Function):#减法
    def forward(self,x0,x1):
        self.x0_shape,self.x1_shape=x0.shape,x1.shape
        return x0-x1
    def backward(self,dout):
        dout0=dout
        dout1=-dout
        if self.x0_shape != self.x1_shape:
            dout0=skystar.utils.sum_to(dout0,self.x0_shape)
            dout1=skystar.utils.sum_to(dout1,self.x1_shape)
        return dout0,dout1
class Div(Function):#除法
    def forward(self,x0,x1):
        self.x0_shape,self.x1_shape=x0.shape,x1.shape
        return x0/x1
    def backward(self,dout):
        x0,x1=self.inputs
        dout1=dout/x1
        dout2=-dout*x0/x1**2
        if self.x0_shape != self.x1_shape:
            dout1=skystar.utils.sum_to(dout1,self.x0_shape)
            dout2=skystar.utils.sum_to(dout2,self.x1_shape)
        return dout1,dout2
class Pow(Function):#幂函数
    def __init__(self,c):
        self.c=c
    def forward(self,x):
        return x**self.c
    def backward(self,dout):
        x=self.inputs[0]
        dout=dout*self.c*x**(self.c-1)
        return dout

def exp(x):
    func=Exp()
    return func(x)
def square(x):
    func=Square()
    return func(x)
def add(x0,x1):
    x1 = as_scalar(x1,skystar.cuda.get_array_module(x0.data))#转换为ndarray，实现Variable与int float的直接运算
    func=Add()
    return func(x0,x1)
def mul(x0,x1):
    x1 = as_scalar(x1,skystar.cuda.get_array_module(x0.data))
    func=Mul()
    return func(x0,x1)
def neg(x):
    func=Neg()
    return func(x)
def sub(x0,x1):
    x1=as_scalar(x1,skystar.cuda.get_array_module(x0.data))
    func=Sub()
    return func(x0,x1)
def rsub(x0,x1):
    x1=as_scalar(x1,skystar.cuda.get_array_module(x0.data))
    func=Sub()
    return func(x1,x0)
def div(x0,x1):
    x1=as_scalar(x1,skystar.cuda.get_array_module(x0.data))
    func=Div()
    return func(x0,x1)
def rdiv(x0,x1):
    x1=as_scalar(x1,skystar.cuda.get_array_module(x0.data))
    func=Div()
    return func(x1,x0)
def pow(x,c):
    func=Pow(c)
    return func(x)

# =============================================================================
'''辅助函数'''
# =============================================================================
class Config:#允许反向传播的开关
    enable_backprop=True
@contextlib.contextmanager
def config_using(name, value):
    oldvalue = getattr(Config, name)
    setattr(Config, name, value)
    try:
        yield
    finally:
        setattr(Config, name, oldvalue)
def no_grad():
    return config_using('enable_backprop', False)
def as_scalar(x,array_module=np):#把数据变为np.ndarray或者cp.ndarray
    if array_module.isscalar(x):
        return array_module.array(x,dtype=array_module.float32)#把数据转为float32
    return x
def as_variable(x):#把数据变为Variable实例
    if isinstance(x,Variable):
        return x
    return Variable(x)

def setup_variable():#重载运算符
    Variable.__mul__=mul
    Variable.__rmul__=mul
    Variable.__add__ = add
    Variable.__radd__ = add
    Variable.__sub__ = sub
    Variable.__rsub__ = rsub
    Variable.__neg__ = neg
    Variable.__truediv__=div
    Variable.__rtruediv__=rdiv
    Variable.__pow__=pow


# =============================================================================
'''一般函数'''
# =============================================================================
'''
x：函数前向传播的inputs，类型为tuple或单个元素，元素类型为Variable,经__call__方法在forward中变为ndarray
y：函数前向传播的outputs，这里输出类型为ndarray，但会在__call__方法中变为Variable
dout：Variabel实例
self.inputs：#list，元素为Variable
self.outputs：#list,元素为Variable实例,使用了弱引用，调用的话需要写为：元素（）
'''
class Sin(Function):
    def forward(self,x):
        xp=skystar.cuda.get_array_module(x)
        y=xp.sin(x)
        return y
    def backward(self,dout):
        x=self.inputs[0]
        dout=dout*cos(x)
        return dout
def sin(x):
    return Sin()(x)
class Cos(Function):
    def forward(self,x):
        xp=skystar.cuda.get_array_module(x)
        y=xp.cos(x)
        return y
    def backward(self,dout):
        x=self.inputs[0]
        dout=dout*-sin(x)
        return dout
def cos(x):
    return Cos()(x)
class Tanh(Function):
    def forward(self,x):
        xp=skystar.cuda.get_array_module(x)
        y=(xp.exp(x)-xp.exp(-x))/(xp.exp(x)+xp.exp(-x))
        return y
    def backward(self,dout):
        y=self.outputs[0]()
        dout=dout*(1-y*y)
        return dout
def tanh(x):
    return Tanh()(x)
class Reshape(Function):
    def forward(self,x,shape):
        self.x_shape=x.shape
        x=x.reshape(shape)
        return x
    def backward(self,dout):
        dout=dout.reshape(self.x_shape)
        return dout
def reshape(x,shape):
    if isinstance(shape,tuple):
        shape=np.asarray(shape)#这里用cupy无法
    shape = Variable(shape,name='shape')
    if x.shape==shape:
        return as_variable(x)
    return Reshape()(x,shape)
class Dot(Function):#点积
    def forward(self,x0,x1):
        xp=skystar.cuda.get_array_module(x0)
        y=xp.dot(x0,x1)
        return y
    def backward(self,dout):
        x0,x1=self.inputs
        dout0=dot(dout,x1.T)
        dout1=dot(x0.T,dout)
        return dout0,dout1
def dot(x0,x1):
    return Dot()(x0,x1)
class Matmul(Function):
    def forward(self,x0,x1):
        xp=skystar.cuda.get_array_module(x0)
        y=xp.matmul(x0,x1)
        return y
    def backward(self,dout):
        x0,x1=self.inputs
        dout0=matmul(dout,x1.data.swapaxes(-1,-2))#这里直接使用numpy或者cupy的swapaxes方法交换最后两轴
        dout1=matmul(x0.data.swapaxes(-1,-2),dout)
        return dout0,dout1
def matmul(x0,x1):
    return Matmul()(x0,x1)
class Transpose(Function):
    def __init__(self, axes=None):
        self.axes = axes
    def forward(self, x):
        y = x.transpose(self.axes)
        return y
    def backward(self, gy):
        if self.axes is None:
            return transpose(gy)
        # xp=skystar.cuda.get_array_module(gy)
        axes_len = len(self.axes)
        inv_axes = tuple(np.argsort([ax % axes_len for ax in self.axes]))
        return transpose(gy, inv_axes)
def transpose(x, axes=None):
    return Transpose(axes)(x)
class Swapaxis(Function):
    def __init__(self, axis1, axis2):
        self.axis1=axis1
        self.axis2=axis2
    def forward(self, x):
        xp=skystar.cuda.get_array_module(x)
        return xp.swapaxes(x,self.axis1,self.axis2)
    def backward(self, dout):
        dx=dout.data.swapaxes(self.axis1,self.axis2)
        dx=Variable(dx)
        return dx
def swapaxis(x, axis1, axis2):
    return Swapaxis(axis1, axis2)(x)
class Sum(Function):
    def __init__(self,axis,keepdims):
        self.axis=axis
        self.keepdims=keepdims
    def forward(self,x):
        self.x_shape=x.shape
        y=x.sum(axis=self.axis,keepdims=self.keepdims)
        return y
    def backward(self,dout):
        dout=skystar.utils.reshape_sum_backward(dout,self.x_shape,self.axis,self.keepdims)
        dout=broadcast_to(dout,self.x_shape)
        return dout
def sum(x,axis=None,keepdims=False):
    return Sum(axis,keepdims)(x)
class BroadcastTo(Function):#广播
    def forward(self,x, shape):
        xp=skystar.cuda.get_array_module(x)
        y=xp.broadcast_to(x,shape)
        return y
    def backward(self,dout):
        input= self.inputs[0]
        dout=sum_to(dout,input.shape)
        return dout
def broadcast_to(x,shape):
    # xp=skystar.cuda.get_array_module(x)
    shape=Variable(np.asarray(shape,dtype=np.int32),name='shape')
    if x.shape==shape:
        return as_variable(x)
    return BroadcastTo()(x,shape)


class Gather(Function):#用于嵌入层
    '''本函数的目的是对数组进行映射，映射有两种基本情况，即一个索引映射一个数据或一行数据
    1、映射一行：输入数据满足（batch,row,col）,索引满足（batch,index,1），axis=1，输出（batch,index,col）
    2、映射一个：输入数据满足（batch,col）,索引满足（batch，1），axis=1，输出（batch，1）
    '''
    def __init__(self, axis=1):
        self.axis = axis  # 默认沿着第 0 轴进行 gather 操作
    def forward(self, x, indices):
        xp=skystar.cuda.get_array_module(x)
        return xp.take_along_axis(x, indices, axis=self.axis)
    def backward(self, dout):
        '''暂时只适用于3D输入[batch,num1,dims],输出[batch,num2,1]，axis==1的情况'''
        xp = skystar.cuda.get_array_module(dout)  # 获取合适的库
        x,indices=self.inputs#Variable
        batch,num,dims=x.shape
        indices=indices.data.squeeze()
        dindices=Variable(xp.zeros_like(indices))
        onehot_indice=xp.eye(num)[indices].transpose(0,2,1)
        dx=xp.matmul(onehot_indice,dout.data)
        dx = Variable(dx)
        return dx,dindices
def gather(x, indices,axis=0):
    '''这里输入的x,indices都是Variable'''
    xp=skystar.cuda.get_array_module(x)
    if not isinstance(indices, Variable):
        indices=Variable(xp.asarray(indices,dtype=xp.int32))
    indices.data=indices.data.astype(xp.int32)
    return Gather(axis)(x, indices)

class SumTo(Function):#广播
    def __init__(self,shape):
        self.shape=shape
    def forward(self,x):
        self.x_shape=x.shape
        y=skystar.utils.sum_to(x,self.shape)
        return y
    def backward(self,dout):
        dout=broadcast_to(dout,self.x_shape)
        return dout
def sum_to(x,shape):
    if x.shape==shape:
        return as_variable(x)
    return SumTo(shape)(x)


class Concat(Function):
    def __init__(self, axis=1):
        self.axis = axis  # 指定拼接的维度轴
    def forward(self, x0, x1):
        # 获取所有输入的形状
        self.shapes = [x0.shape,x1.shape]
        xp=skystar.cuda.get_array_module(x0)
        # 在指定轴上拼接输入
        y = xp.concatenate((x0,x1), axis=self.axis)
        return y
    def backward(self, dout):
        # 根据 forward 时记录的形状信息，将 dout 拆分成与 x0 和 x1 形状相同的部分
        axis = self.axis
        xp=skystar.cuda.get_array_module(dout)
        x0_shape, x1_shape = self.shapes
        # 使用 split 将 dout 分割回 x0 和 x1 的大小
        dx0, dx1 = xp.split(dout.data, [x0_shape[axis]], axis=axis)
        return dx0, dx1
def concat(x0, x1, axis=0):
    return Concat(axis)(x0,x1)
class Slice(Function):
    def __init__(self):
        """
        初始化 Slice 函数。输入可谓list 也可为array
        :param starts: 一个一维张量，包含每个维度的起始点，如 [0, 0, 1, 1]。
        :param ends: 一个一维张量，包含每个维度的结束点，如 [1, 3, 3, 3]。
        :param axis: 一个一维张量，指定 starts 和 ends 应用的轴，如 [0, 1, 2, 3]。
        :param stride: 一个一维张量，指定每个轴的步长，如 [1, 1, 1, 1]。
        """
    def forward(self, x, starts, ends, axis, steps):
        """
        前向计算：对输入张量 x 进行切片。
        :param x: 输入张量。
        :return: 切片后的张量。
        """
        self.input_shape = x.shape  # 保存输入形状以便反向传播使用
        slices = [slice(None)] * len(x.shape)  # 初始化所有维度的切片为 ":" (None)
        for ax, start, end, step in zip(axis, starts, ends, steps):
            slices[ax] = slice(start, end, step)  # 按指定轴应用切片
        self.slices = slices  # 保存切片信息供反向传播使用
        y = x[tuple(slices)]  # 应用切片
        return y

    def backward(self, dout):
        """
        反向传播：将 dout 放回原来的形状，并在非切片部分填充 0。
        :param dout: 梯度流。
        :return: 对输入 x 的梯度。
        """
        xp = skystar.cuda.get_array_module(dout)
        dx = xp.zeros(self.input_shape, dtype=dout.dtype)  # 初始化为零的张量，与输入形状相同
        dx[tuple(self.slices)] = dout  # 将 dout 放回切片位置
        return dx

def my_slice(x, starts, ends, axis=None, steps=None):
    if axis is None:
        axis = list(range(len(starts)))  # 默认为所有轴
    if steps is None:
        steps = [1] * len(starts)  # 默认为步长为 1
    xp=skystar.cuda.get_array_module(x)
    starts = Variable(xp.array(starts,dtype=xp.int32),name='starts')
    ends = Variable(xp.array(ends,dtype=xp.int32),name='ends')
    steps = Variable(xp.array(steps,dtype=xp.int32),name='steps')
    axis = Variable(xp.array(axis,dtype=xp.int32),name='axis')
    return Slice()(x,starts, ends, axis, steps)
class Mean(Function):
    def __init__(self, axis=None, keepdims=True):
        self.axis = axis  # 支持传入 axis 参数
        self.keepdims=keepdims
    def forward(self, x):
        xp=skystar.cuda.get_array_module(x)
        self.original_x_shape = x.shape  # 保存原始输入形状
        self.x = x
        # 计算均值
        if self.axis is None:
            mean = xp.mean(x,keepdims=self.keepdims)  # 没有 axis 参数时，计算整个张量的均值
        else:
            mean= xp.mean(x, axis=self.axis, keepdims=self.keepdims)  # 按照指定的 axis 计算均值

        return mean

    def backward(self, dout):
        xp=skystar.cuda.get_array_module(dout)
        N = xp.prod(self.original_x_shape,dtype=xp.float32)  # 计算输入张量的总元素数
        axis_len = xp.prod(xp.array(self.original_x_shape)[self.axis],dtype=xp.float32) if self.axis is not None else N
        # 这里如何不是keepdims=True，则会报错，默认keepdims=True
        dx = xp.ones_like(self.x,dtype=xp.float32) * dout.data / axis_len  # 使用 np.ones_like 保证 dx 形状与 x 相同
        return dx
def my_mean(x, axis=None,keepdims=True):
    return Mean(axis,keepdims=keepdims)(x)


# =============================================================================
'''激活函数'''
# =============================================================================
class Sigmoid(Function):
    def forward(self,x):
        xp=skystar.cuda.get_array_module(x)
        y=1/(1+xp.exp(-x))
        return y
    def backward(self,dout):
        y=self.outputs[0]()
        dout = dout * y * (1 - y)
        return dout
def sigmoid(x):
    func=Sigmoid()
    return func(x)

class Relu(Function):
    def __init__(self):
        self.mask=None
    def forward(self,x):
        self.mask=(x<=0)#标记小于等于0的位置
        y=x.copy()#制作一个副本，避免x被改变
        y[self.mask]=0#小于等于0的值输出0，其它数值原样输出
        return y
    def backward(self,dout):
        dout.data[self.mask]=0
        return dout
def ReLU(x):
    return Relu()(x)

class Softmax(Function):#用于分类的激活函数,用于输出和输入相等
    def __init__(self,axis=None):
        if axis is None:
            self.axis=1
        else:
            self.axis=axis
    def forward(self,x):
        xp = skystar.utils.get_array_module(x)
        c = xp.max(x, axis=self.axis,keepdims=True)  # 求出最大值，避免数据溢出
        x_exp = xp.exp(x - c)  # 利用了广播，每一列的样本减去相同的值,溢出对策
        sum = xp.sum(x_exp, axis=self.axis,keepdims=True)
        y = x_exp / sum
        return y
    def backward(self,dout):
        y=self.outputs[0]()
        return dout*y*(1-y)
def softmax(x,axis=None):
    return Softmax(axis)(x)
# =============================================================================
'''loss函数'''
# =============================================================================
class SoftmaxCrossEntropyLoss(Function):#该函数结合了softmax和loss交叉熵误差函数，可以用作分类问题的最后一个激活函数
    '''y[batch,num_classes,H,W],[batch,index,num_classes],[batch,num_classes]
    t[batch,1,H,W],[batch,index],[batch,]'''
    def __init__(self,axis=None):
        if axis is not None:
            self.axis=axis
        else:self.axis=-1
    def forward(self,x0,x1):
        y=softmax(x0,axis=self.axis)
        loss=skystar.utils.cross_entropy_error(y.data,x1)
        return loss
    def backward(self,dout):#输出层反向传播输入导数1
        x0,t=self.inputs
        y=softmax(x0,axis=self.axis)#Variable
        if t.ndim==1:
            batch,num_class=y.shape
            t=Variable(skystar.utils.onehot(t.data,num_class))#Variable
            dout = (y - t) / batch
        elif t.ndim==2:
            batch, index, num_class = y.shape
            t=Variable(skystar.utils.onehot(t.data,num_class))#Variable
            dout = (y - t) / batch
        elif t.ndim==4:
            t.data=t.data.squeeze(axis=1)
            batch,num_class,H,W = y.shape
            t=Variable(skystar.utils.onehot(t.data,num_class).transpose(0,3,1,2))#Variable
            dout = (y - t) / batch
        else:
            raise ValueError('t.ndim==1 or t.ndim==2 or t.ndim==4')
        return dout
def softmaxwithloss(x,t,axis=None):
    '''t:np.array或者cp.array'''
    xp=skystar.utils.get_array_module(x)
    t=Variable(xp.asarray(t),name='Label')#给t添加名字
    return SoftmaxCrossEntropyLoss(axis)(x, t)

class MeanSquaredError(Function):
    "均方误差，多用于时序模型和一些需要精准预测结果的模型"
    def forward(self,x0,x1):
        diff=x0-x1
        y=(diff**2).sum()/len(diff)
        return y
    def backward(self,dout):
        x0,x1=self.inputs
        diff=x0-x1
        gx0=dout*diff*(2./len(diff))
        gx1=-gx0
        return gx0,gx1
def mean_squared_error(x0,x1):
    return MeanSquaredError()(x0,x1)

# =============================================================================
'''层函数'''
# =============================================================================
class Gemm(Function):
    '''
    也叫Affine层
    affine只进行二维矩阵的点积
    x：输入数据，如果输入数据是多维，则先变为2维
    W：参数权重，为二维的Varaible
    b，参数偏置，为一维的Variable
    transA: 是否对x进行转置
    transB：是否对权重w进行转置
    affine层可以单纯由dot与add实现
    '''
    def __init__(self,alpha=1,beta=1,transA=False,transB=False):
        self.alpha=alpha
        self.beta=beta
        self.transA=transA
        self.transB=transB
    def forward(self,x,W,b):
        xp=skystar.cuda.get_array_module(x)
        self.original_x_shape=x.shape
        if self.transA:
            x=x.T
        if self.transB:
            W=W.T
        if b is None:
            y = xp.dot(x, W) * self.alpha
        else:
            y = xp.dot(x, W) * self.alpha + b * self.beta
        return y
    def backward(self,dout):
        x,W,b=self.inputs
        dW=dot(x.T,dout) * self.alpha
        if b is not None:
            db = sum(dout, axis=0) * self.beta
        dout=dot(dout,W.T) * self.alpha
        dout=dout.reshape(*self.original_x_shape)
        if b is None:
            return dout,dW
        else:
            return dout,dW,db
def gemm(x, W, b, alpha=1, beta=1, transA=False, transB=False):
    return Gemm(alpha,beta,transA,transB)(x,W,b)
# =============================================================================
'''BatchNormalization'''
# =============================================================================
class BatchNormalization(Function):
    '''batchnorm 均值和方差是针对特征通道计算的，对于二维数据（N,B），特征通道是B，对于（N,C,H,W），特征通道是C,暂时不考虑3D数据'''
    def __init__(self,momentum=0.9,eps=1e-5):
        self.momentum=momentum#float
        self.xv=None#ndarray
        self.var=None#ndarray
        self.mean=None#ndarray

        self.test_mean=None#ndarray
        self.test_var=None#ndarray
        self.eps=eps
    def forward(self,x,gamma,beta,input_mean,input_var):
        xp=skystar.cuda.get_array_module(x)
        self.test_mean=input_mean
        self.test_var=input_var
        if Get_TrainingMode():#训练时用当前批量输入的均值方差
            if x.ndim==2:
                sum = xp.sum(x, axis=0)
                N = x.shape[0]
                mean = sum / N
                var = xp.sum((x - mean) ** 2, axis=0) / N
            elif x.ndim==4:
                sum = xp.sum(x, axis=(0,2,3))
                N=x.shape[0]*x.shape[2]*x.shape[3]
                mean = sum / N
                var = xp.sum((x - mean.reshape(mean.shape[0],1,1)) ** 2, axis=(0,2,3)) / N
            else:
                raise IndexError
            self.mean = mean
            self.var = var
            '''更新全局均值方差，用于测试用'''
            self.test_mean=self.momentum * self.test_mean + (1 - self.momentum) * self.mean
            self.test_var=self.momentum * self.test_var + (1 - self.momentum) * self.var
        else:#测试用全局均值方差
            mean=self.test_mean
            var=self.test_var
        if x.ndim==4:
            x=x.transpose(0,2,3,1)
            self.xv = ((x - mean) / xp.sqrt(var + self.eps))
            y = gamma * self.xv + beta

            self.xv=self.xv.transpose(0,3,1,2)
            y=y.transpose(0,3,1,2)
            return y
        if x.ndim==2:
            self.xv = (x - mean) / xp.sqrt(var + self.eps)
            y = gamma * self.xv + beta
            return y
    def backward(self,dout):
        xp=skystar.cuda.get_array_module(dout)
        x,gamma,beta=self.inputs[0],self.inputs[1],self.inputs[2]#Variable
        if x.ndim==2:
            N=dout.shape[0]#ndarray
            dgamma=sum(dout*self.xv,axis=0)#Variable
            dbeta=sum(dout,axis=0)#Variable
            dxv=dout*gamma#Variable
            dvar=sum(dxv*(x-self.mean)*xp.power(self.var + self.eps, -1.5)*(-0.5),axis=0)#Variable
            dmean=-1.0*sum(dxv,axis=0)/xp.sqrt(self.var+self.eps)+dvar*sum(-2.0*(x-self.mean),axis=0)/N#Variable
            dx=dxv/xp.sqrt(self.var+self.eps)+dvar*2*(x-self.mean)/N+dmean/N#Variable
        else:
            N=dout.shape[0]*dout.shape[2]*dout.shape[3]
            dgamma=sum(dout*self.xv,axis=(0,2,3))
            dbeta=sum(dout,axis=(0,2,3))

            dout=dout.transpose(0,2,3,1)
            x=x.transpose(0,2,3,1)
            dxv=dout*gamma

            dvar=sum(dxv*(x-self.mean)*xp.power(self.var + self.eps, -1.5)*(-0.5),axis=(0,1,2))#Variable
            dmean=-1.0*sum(dxv,axis=(0,1,2))/xp.sqrt(self.var+self.eps)+dvar*sum(-2.0*(x-self.mean),axis=(0,1,2))/N#Variable
            dx=dxv/xp.sqrt(self.var+self.eps)+dvar*2*(x-self.mean)/N+dmean/N#Variable
            dx=dx.transpose(0,3,1,2)
        return dx,dgamma,dbeta
# =============================================================================
'''LayerNorm'''
# =============================================================================
class LayerNorm(Function):
    '''只考虑了3D输入的情况（batch，seqlen，embedding_nidm）'''
    def __init__(self, epsilon=1e-5):
        self.epsilon = epsilon  # 防止除零的小常数

    def forward(self, x, gamma, beta):
        xp=skystar.cuda.get_array_module(x)
        # 计算均值和方差,对最后一维进行归一化，这里与batchnorm不同
        self.mean = xp.mean(x, axis=-1, keepdims=True)
        self.var = xp.var(x, axis=-1, keepdims=True)
        # 归一化
        self.x_hat = (x - self.mean) / xp.sqrt(self.var + self.epsilon)
        # 缩放和平移
        out = gamma * self.x_hat + beta
        return out

    def backward(self, dout):
        xp=skystar.cuda.get_array_module(dout)
        x,gamma,beta=self.inputs
        # 计算梯度
        Bath,N, D = dout.shape  # 输入的样本数量和特征维度
        dx_hat = dout * gamma  # 先计算归一化的梯度
        # 计算方差和均值的梯度
        dvar = xp.sum(dx_hat.data * (self.x_hat * -0.5) * xp.power(self.var + self.epsilon, -1.5), axis=-1, keepdims=True)
        dmean = xp.sum(dx_hat.data * -1 / xp.sqrt(self.var + self.epsilon), axis=-1, keepdims=True) + dvar * xp.mean(-2 * self.x_hat, axis=-1, keepdims=True)
        # 计算输入的梯度
        dx = Variable(dx_hat.data / xp.sqrt(self.var + self.epsilon) + dvar * 2 * self.x_hat / D + dmean / D)
        dgamma = Variable(xp.sum(dout.data * self.x_hat, axis=(0,1)))#
        dbeta = Variable(xp.sum(dout.data, axis=(0,1)))#
        # 返回梯度
        return dx, dgamma, dbeta
def layernorm(x,gamma,beta,epsilon=1e-5):
    return LayerNorm(epsilon)(x,gamma,beta)
# =============================================================================
'''Dropout'''
# =============================================================================
class Dropout(Function):
    '''训练时随机删除神经元节点，有效避免过拟合'''
    def __init__(self):
        self.mask=None
    def forward(self,x, ratio, training_mode):
        xp=skystar.cuda.get_array_module(x)
        if Get_TrainingMode():
            rand = xp.random.rand(*x.shape)  # 生成与x形状相同的随机数
            self.mask = (rand < ratio)
            y = x.copy()
            y[self.mask] = 0#被标记的神经元
            y/=(1.0 - ratio)#缩放处理，保持输出一致性
        else:
            y=x#测试时不对数据进行缩放，因为所有神经元都要输出
        return y
    def backward(self, dout):
         dout.data[self.mask]=0
         return dout
def dropout(x,ratio=0.5,training_mode=False):
    xp=skystar.cuda.get_array_module(x)
    ratio = Variable(xp.array([ratio],dtype=xp.float32),name='ratio')
    training_mode = Variable(xp.array([Get_TrainingMode()],dtype=xp.int32),name='training_mode')
    return Dropout()(x,ratio,training_mode)

# =============================================================================
'''卷积网络层函数'''
# =============================================================================
class Conv(Function):
    '''
    卷积层
    W：Variable,四维shape(out_channel,in_channel,FH,FW)
    B：Variable,一维shape(out_channel,)
    stride:步长
    pad:填充
    to_affine:如果下一层是affine层，将输出调整为矩阵，而不是四维张量
    '''
    def __init__(self,stride=1,pad=0):
        self.stride=stride#int
        self.pad=pad#int

        # 中间数据（backward时使用）
        self.col = None#array,shape(N*oh*ow, c*fh*fw)
        self.col_W = None#array,shape(c*fh*fw,FN)
    def forward(self,x,Weight,b):
        xp=skystar.cuda.get_array_module(x)
        N,C,H,W=x.shape
        FN,C,FH,FW=Weight.shape
        out_h=int((H+2*self.pad-FH)//self.stride+1)
        out_w=int((W+2*self.pad-FW)//self.stride+1)
        self.col=skystar.utils.im2col(x,FH,FW,stride=self.stride,pad=self.pad)#shape(N*oh*ow, c*fh*fw)
        self.col_W=Weight.reshape(FN,-1).T#shape(c*fh*fw,FN)

        out = xp.dot(self.col, self.col_W)
        if b is not None:
            out +=b  # shape(N*oh*ow,FN))
        out=out.reshape(N,out_h,out_w,-1).transpose(0,3,1,2)#shape(N,FN,oh,ow)
        return out

    def backward(self,dout):
        x,W,b=self.inputs[0],self.inputs[1],self.inputs[2]
        FN, C, FH, FW = W.shape
        dout = dout.transpose(0, 2, 3, 1).reshape(-1, FN)#(N*oh*ow,FN)

        dw = dot(self.col.T, dout).transpose(1,0).reshape(FN,C,FH,FW)

        dcol = dot(dout, self.col_W.T)#Variable(N*oh*ow,c*fh*fw)
        dx = skystar.utils.col2im(dcol.data, x.shape, FH, FW, self.stride, self.pad)#shape(N,C,H,W)
        dx=Variable(dx)
        if b is not None:
            db = sum(dout, axis=0)  # (FN,)
            return dx, dw, db
        return dx, dw
def convolution(x,W,b,stride=1,pad=0):
    return Conv(stride,pad)(x,W,b)

class ConvTranspose(Function):
    def __init__(self,stride=1,pad=0):
        self.stride=stride#int
        self.pad=pad#int

        # 中间数据（backward时使用）
        self.col = None#array,shape(N*oh*ow, c*fh*fw)
        self.col_W = None#array,shape(c*fh*fw,FN)
    def forward(self,x,Weight,b):
        '''2*2-->填充k=3,s=2,p=0--> 7*7-->标准卷积k=3,s=1,p=0-->5*5
        o' = s(i' - 1) + k-2p ，i'为原输入尺寸，o'为输出尺寸 （s=32,p=16,k=64）适合图像尺寸224*224卷积为7*7的网络反卷积'''
        '''注意，ConvTranspose的权重形状为（in_channels,out_channels,H,W）
        输入为（N,in_channels,H，W）,因此需要对权重进行轴的调换'''
        xp=skystar.cuda.get_array_module(x)#3*3
        Weight=Weight.transpose(1,0,2,3)
        out_channel, in_channel, FH, FW = Weight.shape
        reWeight = xp.flip(Weight, axis=(2, 3))  # 对权重矩阵进行翻转
        x=skystar.utils.transconv_pad(x,stride=self.stride,kernel_size=FH,pad=self.pad)#对x进行反卷积的填充，7*7
        N, C, H, W = x.shape
        self.shape=x.shape

        #标准卷积
        out_h=int((H+2*0-FH)//1+1)#5
        out_w=int((W+2*0-FW)//1+1)#5
        self.col=skystar.utils.im2col(x,FH,FW,stride=1,pad=0)#shape(N*oh*ow, c*fh*fw)
        self.col_W=reWeight.reshape(out_channel,-1).T#shape(c*fh*fw,out_channel)
        out=xp.dot(self.col, self.col_W)
        if b is not None:
            out += b  # shape(N*oh*ow,FN)
        out=out.reshape(N,out_h,out_w,-1).transpose(0,3,1,2)#shape(N,out_channel,oh,ow)5*5

        return out
    def backward(self,dout):
        x,W,b=self.inputs[0],self.inputs[1],self.inputs[2]#这里的W是Variable实例
        xp = skystar.cuda.get_array_module(x)
        W=W.transpose(1,0,2,3)
        out_channel, in_channel, FH, FW = W.shape
        dout = dout.transpose(0, 2, 3, 1).reshape(-1, out_channel)#(N*oh*ow,out_channel)

        dw = dot(self.col.T, dout).transpose(1, 0).reshape(out_channel, in_channel, FH, FW)#这里得到的梯度是翻转后W的梯度，因此梯度需要翻转
        dw = Variable(xp.flip(dw.data.transpose(1, 0, 2, 3), axis=(2, 3)))

        dcol = dot(dout, self.col_W.T)#Variable(N*oh*ow,c*fh*fw)
        dx = skystar.utils.col2im(dcol.data, self.shape, FH, FW, stride=1,pad=0)#shape(N,C,H,W)
        dx=Variable(skystar.utils.back_transcov_pad(dx,stride=self.stride,kernel_size=FH,pad=self.pad))
        if b is not None:
            db = sum(dout, axis=0)  # (FN,)
            return dx, dw, db
        return dx,dw
def transposed_convolution(x,weight,b,stride=1,pad=0):
    return ConvTranspose(stride=stride, pad=pad)(x, weight, b)

class MaxPool(Function):
    '''池化层，取池化窗口的最大值'''
    def __init__(self, pool_size, stride=1, pad=0):
        self.pool_size = pool_size#int
        self.stride = stride#int
        self.pad = pad#int
        self.arg_max = None#array

    def forward(self, x):
        xp=skystar.cuda.get_array_module(x)
        N, C, H, W = x.shape
        out_h = int(1 + (H - self.pool_size) / self.stride)
        out_w = int(1 + (W - self.pool_size) / self.stride)

        '''展开为二维数组'''
        col = skystar.utils.im2col(x, self.pool_size, self.pool_size, self.stride, self.pad)
        col = col.reshape(-1, self.pool_size * self.pool_size)

        '''计算最大值'''
        out = xp.max(col, axis=1)
        self.arg_max = xp.argmax(col, axis=1)  # 记录最大值的位置
        '''转换为四维数组'''
        out = out.reshape(N, out_h, out_w, C).transpose(0, 3, 1, 2)

        return out

    def backward(self, dout):
        xp=skystar.cuda.get_array_module(dout)
        x = self.inputs[0]
        dout = dout.transpose(0, 2, 3, 1)

        pool_size = self.pool_size * self.pool_size
        dmax = xp.zeros((dout.size, pool_size))
        dmax[xp.arange(self.arg_max.size), self.arg_max.flatten()] = dout.data.flatten()
        dmax = dmax.reshape(dout.shape + (pool_size,))

        dcol = dmax.reshape(dmax.shape[0] * dmax.shape[1] * dmax.shape[2], -1)
        dx = skystar.utils.col2im(dcol, x.shape, self.pool_size, self.pool_size, self.stride, self.pad)
        dx = Variable(dx)
        return dx
def maxpool(x,pool_size,stride=1,pad=0):
    return MaxPool(pool_size,stride,pad)(x)

class AveragePool(Function):
    '''平均池化层，取池化窗口的平均值'''
    def __init__(self, pool_size, stride=1, pad=0):
        self.pool_size = pool_size  # int
        self.stride = stride  # int
        self.pad = pad  # int
    def forward(self, x):
        xp = skystar.cuda.get_array_module(x)
        N, C, H, W = x.shape
        out_h = int(1 + (H - self.pool_size) / self.stride)
        out_w = int(1 + (W - self.pool_size) / self.stride)
        # 展开为二维数组
        col = skystar.utils.im2col(x, self.pool_size, self.pool_size, self.stride, self.pad)
        col = col.reshape(-1, self.pool_size * self.pool_size)
        # 计算平均值
        out = xp.mean(col, axis=1)
        # 转换为四维数组
        out = out.reshape(N, out_h, out_w, C).transpose(0, 3, 1, 2)
        return out
    def backward(self, dout):
        x = self.inputs[0]
        dout = dout.transpose(0, 2, 3, 1)
        # 计算反向传播
        pool_size = self.pool_size * self.pool_size
        dmean = dout / pool_size
        dmean = Variable(dmean.data.repeat(pool_size, axis=-1))

        dcol = dmean.reshape(dmean.shape[0] * dmean.shape[1] * dmean.shape[2], -1)
        dx = skystar.utils.col2im(dcol.data, x.shape, self.pool_size, self.pool_size, self.stride, self.pad)
        dx = Variable(dx)
        return dx
def avgpool(x,pool_size,stride=1,pad=0):
    return AveragePool(pool_size,stride,pad)(x)

# class Gather(Function):
#     def __init__(self, shape):
#         pass
#     def forward(self, x):
#         pass




import setuptools  # 导入setuptools打包工具

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# 读取 requirements.txt 中的依赖
with open("requirements.txt", "r",encoding="utf-8") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="skystar",
    version="1.0.55",
    author="Yonas-xin",
    author_email="linkstar443@163.com",
    description="A Deep Learning Framework For Beginners",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Yonas-Xin/STARSKY",
    packages=['skystar', 'skystar.sky_dataset'],
    install_requires=requirements,  # 从 requirements.txt 读取依赖
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',  # Minimum Python version required

    entry_points={
        'console_scripts': [
            'skystar = skystar.__main__:main'
        ]}
)



import numpy as np

class Optimizer:
    '''
    target:目标model
    hooks:储存预处理功能函数
    '''
    def __init__(self,use_iter=False):
        self.use_iter=use_iter
        self.iter=0
        self.target=None
        self.hooks=[]
    '''target为model'''
    def setup(self,target):
        self.target=target
        return self

    def to_iter(self):
        self.use_iter=True

    def update(self):
        params=[param for param in self.target.params() if param.grad is not None]

        #可选功能，预处理，比如权重衰减
        for f in self.hooks:
            f(params)

        if not self.use_iter:
            for param in params:
                self.update_one(param)
        else:
            self.iter+=1
            for param in params:
                self.update_one(param,self.iter)

    def update_one(self,param,iter=0):
        raise NotImplementedError

    def add_hook(self,f):#使用该功能添加预处理方法
        self.hooks.append(f)


class SGD(Optimizer):
    '''这种更新方法学习率保持不变'''
    def __init__(self,lr=0.01):
        super().__init__()
        self.lr=lr

    def update_one(self,param):
        param.data-=self.lr*param.grad.data

class MomentumSGD(Optimizer):
    '''这种更新方法所有参数学习率一起变化，逐渐减小，直到稳定'''
    def __init__(self,lr=0.01,momentum=0.9):
        super().__init__()
        self.lr=lr
        self.momentum=momentum
        self.vs={}

    def update_one(self,param):
        v_key=id(param)
        if v_key not in self.vs:
            self.vs[v_key]=np.zeros_like(param.data)

        v=self.momentum*self.vs[v_key]-self.lr*param.grad.data
        param.data+=v


class AdaGrad(Optimizer):
    '''这种更新方法每个参数的学习率的变化速度不一样'''
    def __init__(self,lr=0.01):
        super().__init__()
        self.lr=lr
        self.hs={}

    def update_one(self,param):
        h_key=id(param)
        if h_key not in self.hs:
            self.hs[h_key]=np.zeros_like(param.data)

        self.hs[h_key]+=param.grad.data*param.grad.data
        param.data-=self.lr*param.grad.data/(np.sqrt(self.hs[h_key]+1e-6))


class Adam(Optimizer):
    """这种更新方法集合了momentum法与Adagrad的优点。Adam (http://arxiv.org/abs/1412.6980v8)"""

    def __init__(self, lr=0.001, beta1=0.9, beta2=0.999):
        super(Adam, self).__init__(use_iter=True)
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.iter = 0
        self.ms = {}
        self.vs = {}

    def update_one(self, param, iter=0):
        _key=id(param)
        if _key not in self.ms:
            self.ms[_key]=np.zeros_like(param.data)
            self.vs[_key]=np.zeros_like(param.data)

        lr_t = self.lr * np.sqrt(1.0 - self.beta2 ** iter) / (1.0 - self.beta1 ** iter)
        self.ms[_key] += (1 - self.beta1) * (param.grad.data - self.ms[_key])
        self.vs[_key] += (1 - self.beta2) * (param.grad.data ** 2 - self.vs[_key])
        param.data -= lr_t * self.ms[_key] / (np.sqrt(self.vs[_key]) + 6)



'''Info:该脚本用于制作Voc2012的数据集，但是删除了matplotlib的接口，后续考虑使用PIL接口'''
# from matplotlib import image
import os
import numpy as np

'''颜色条'''
VOC_COLORMAP = [[245, 245, 245], [128, 0, 0], [0, 128, 0], [128, 128, 0],
                [0, 0, 128], [128, 0, 128], [0, 128, 128], [128, 128, 128],
                [64, 0, 0], [192, 0, 0], [64, 128, 0], [192, 128, 0],
                [64, 0, 128], [192, 0, 128], [64, 128, 128], [192, 128, 128],
                [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0],
                [0, 64, 128]]
'''类别'''
VOC_CLASSES = ['background', 'aeroplane', 'bicycle', 'bird', 'boat',
               'bottle', 'bus', 'car', 'cat', 'chair', 'cow',
               'diningtable', 'dog', 'horse', 'motorbike', 'person',
               'potted plant', 'sheep', 'sofa', 'train', 'tv/monitor']
def pad(feature,label,height,width):
    '''如果图像不支持切割成预定的大小，则对图像进行扩充'''
    H,W,C=feature.shape
    if H<=height:
        pad=height-H
        pad+=10#上下加10个像素
        feature=np.pad(feature,[(pad,pad),(0,0),(0,0)],mode='constant',constant_values=0)
        try:
            label = np.pad(label, [(pad, pad), (0, 0), (0, 0)], mode='constant', constant_values=0)
        except ValueError:
            # plt.imshow(label)
            # plt.show()
            pass
            label = np.pad(label, [(pad, pad), (0, 0)], mode='constant', constant_values=0)
    if W<=width:
        pad=width-W
        pad+=10
        feature=np.pad(feature,[(0,0),(pad,pad),(0,0)],mode='constant',constant_values=0)
        label=np.pad(label,[(0,0),(pad,pad),(0,0)],mode='constant',constant_values=0)
    return feature,label
def read_voc_images(voc_dir, is_train=True):
    """
    读取所有VOC图像并标注
    :param voc_dir: 储存voc官方数据的根目录
    :param is_train: bool，Ture读取训练数据，False读取测试数据
    :return: list，元素为ndarry
    """
    txt_fname = os.path.join(voc_dir, 'ImageSets', 'Segmentation',
                             'train.txt' if is_train else 'val.txt')
    with open(txt_fname, 'r') as f:
        images = f.read().split()
    features, labels = [], []
    for i, fname in enumerate(images):
        pass
        # features.append(image.imread(os.path.join(
        #     voc_dir, 'JPEGImages', f'{fname}.jpg')))
        # labels.append((image.imread(os.path.join(
        #     voc_dir, 'SegmentationClass', f'{fname}.png'))*255).astype(np.uint8))
    return features, labels
#@save
def voc_colormap2label():
    '''
    构建从RGB到VOC类别索引的映射
    :return: 返回一维矩阵，如RGB=1，2，3 对应索引为1 则colormap2label[(1*256+2)*256+3]=1
    '''
    colormap2label = np.zeros(256**3)
    for i, colormap in enumerate(VOC_COLORMAP):
        index=(colormap[0] * 256 + colormap[1]) * 256 + colormap[2]
        colormap2label[index] = i
    return colormap2label

def voc_label_indices(colormap, colormap2label):
    '''
    将VOC标签中的RGB值映射到它们的类别索引
    :param colormap: 输入图像，三通道标签图shape（H，W，C）
    :param colormap2label: RGB到索引的映射矩阵，一维ndarray
    :return: 单通道标签图shape（H，W）
    '''
    colormap = colormap.astype(np.float32)
    idx = ((colormap[:, :, 0] * 256 + colormap[:, :, 1]) * 256
           + colormap[:, :, 2])
    idx = idx.astype(np.int32)
    # idx:目标矩阵，元素值为colormap2label的索引 #colormap2label：映射矩阵，元素值为标签
    return colormap2label[idx]
def rand_crop(orijin_feature, orijin_label, height, width,num):
    '''
    :param orijin_feature: 输入图像，ndarray，shape(height, width, c)
    :param orijin_label: 输入标签，ndarray，shape(height, width, c)
    :param height: 裁剪高度
    :param width: 裁剪宽度
    :param num: 裁剪数量
    :return: ndarray shape（num，height，width，c）
    '''
    orijin_feature,orijin_label=pad(orijin_feature,orijin_label,height,width)#这里如果图像大小不符合，则对图像进行扩充
    features,labels=[],[]
    for i in range(num):
        feature, label = voc_rand_crop(orijin_feature, orijin_label, height, width)
        features.append(feature)
        labels.append(label)
    features=np.array(features)
    labels=np.array(labels)
    return features, labels

def dataset_rand_crop(feature_dataset, label_dataset1, height, width,mutify=10):
    '''
    :param feature_dataset: 输入图像，list，元素为ndarray
    :param label_dataset1: 输入标签，list，元素为ndarray
    :param height: 裁剪高度
    :param width: 裁剪宽度
    :param mutify: 扩充倍数
    :return: ndarray，四维矩阵
    '''
    x_data,t_data=[],[]
    for i in range(len(feature_dataset)):
        features,labels=rand_crop(feature_dataset[i], label_dataset1[i], height, width,num=mutify)
        x_data.append(features)
        t_data.append(labels)
    x_data=np.array(x_data)
    num1,num2,H,W,C=x_data.shape
    x_data=x_data.reshape(num1*num2,H,W,C)
    t_data=np.array(t_data)
    num1,num2,H,W,C=t_data.shape
    t_data=t_data.reshape(num1*num2,H,W,C)
    return x_data,t_data
def voc_rand_crop(feature, label, height, width):
    '''裁剪图像和标签图像'''
    def random_crop(feature, crop_size):
        """对输入的图像进行随机裁剪，返回裁剪后的图像及裁剪区域"""
        crop_width, crop_height = crop_size
        height, width = feature.shape[:2]
        top = np.random.randint(0, height - crop_height + 1)
        left = np.random.randint(0, width - crop_width + 1)
        cropped_feature = feature[top:top + crop_height, left:left + crop_width]
        return cropped_feature, (left, top, crop_width, crop_height)
    def fixed_crop(image, left, top, crop_width, crop_height):
        """从输入的图像中根据给定的矩形区域裁剪出图像"""
        return image[top:top + crop_height, left:left + crop_width]
    """随机裁剪特征和标签图像"""
    feature, rect = random_crop(feature, (width, height))
    label = fixed_crop(label, *rect)
    return feature, label
def create_data(path, is_train=True,height=224,width=224,mutify=10):
    '''
    :param path: 存放数据的根目录
    :param is_train: 训练数据or测试数据
    :param height: 裁剪高度
    :param width: 裁剪宽度
    :param mutify: 扩充倍数
    :return: 数据集，四维
    '''
    feature,label=read_voc_images(path, is_train)#从根目录获取图像数据，list
    feature,label=dataset_rand_crop(feature,label,height,width,mutify=mutify)#随机裁剪，扩充图像数量，ndarray
    t=[]
    for i in range(len(label)):
        t.append(voc_label_indices(label[i],voc_colormap2label()))
    t=np.array(t).astype(np.uint8)#三通道标签数据转为单通道标签数据
    return feature,t
#请指定VOC2012数据的绝对位置，如下所示
vocdir='D:\\Programing\\pythonProject\\Dezero\\voc2012_dataset\\VOC2012'

# feature1,label1=create_data(vocdir,is_train=True,mutify=10)
# feature2,label2=create_data(vocdir,is_train=False,mutify=1)
# np.savez_compressed('voc2012_train.npz',feature=feature1,label=label1)
# np.savez_compressed('voc2012_test.npz',feature=feature2,label=label2)


import numpy as np
gpu_enable=True
try:#试着能否导入cupy
    import cupy as cp
    cupy=cp
    # print('GPU enabled')
except ImportError:
    gpu_enable=False
from skystar.core import Variable

def get_array_module(x):#获取数据的模块，既numpy 或者 cupy
    if isinstance(x,Variable):
        x=x.data
    if not gpu_enable:
        return np
    xp=cp.get_array_module(x)
    return xp

def as_numpy(x):#把数据转化为numpy.ndarray
    if isinstance(x,Variable):
        x=x.data

    if np.isscalar(x):
        return np.array(x)
    elif isinstance(x,np.ndarray):
        return x
    return cp.asnumpy(x)

def as_cupy(x):#把数据转化为cupy.ndarray
    if isinstance(x,Variable):
        x=x.data

    if not gpu_enable:
        raise Exception('CuPy cannot be loaded.Install CuPy')
    return cp.asarray(x)